# test_annotate.py
# Contact: Jacob Schreiber <jmschreiber91@gmail.com>


import torch
torch.use_deterministic_algorithms(True, warn_only=True)
torch.manual_seed(0)

import numpy
import pytest

from numpy.testing import assert_raises
from numpy.testing import assert_array_almost_equal

from tangermeme.utils import one_hot_encode
from tangermeme.utils import random_one_hot

from tangermeme.seqlet import recursive_seqlets

from tangermeme.annotate import annotate_seqlets
from tangermeme.annotate import count_annotations
from tangermeme.annotate import pairwise_annotations


@pytest.fixture
def annotations():
	r = numpy.random.RandomState(1)
	
	y = torch.zeros(100, 2, dtype=torch.int64)
	y[:, 0] = torch.from_numpy(r.choice(25, size=100))
	y[:, 1] = torch.from_numpy(r.choice(7, size=100))
	return y


@pytest.fixture
def X():
	return torch.load("tests/data/X_gata.torch")


@pytest.fixture
def X_contrib(X):
	X_attr = torch.load("tests/data/X_gata_attr.torch")
	return (X * X_attr).sum(axis=1)


###


def test_annotate_seqlets(X, X_contrib):
	seqlets = recursive_seqlets(X_contrib)
	assert seqlets.shape == (24, 5)

	idxs, pvals = annotate_seqlets(X, seqlets, "tests/data/test.meme")

	assert idxs.shape == (24, 1)
	assert idxs.dtype == torch.int32

	assert pvals.shape == (24, 1)
	assert pvals.dtype == torch.float64

	assert_array_almost_equal(idxs[:, 0], [
		9, 6, 9, 3, 6, 9, 9, 9, 6, 6, 9, 8, 9, 9, 7, 7, 6, 9, 6, 7, 9, 8, 6, 7
	])

	assert_array_almost_equal(pvals[:, 0], [
		0.0641, 0.0117, 0.0053, 0.0923, 0.0178, 0.0196, 0.0803, 0.0341, 0.0178, 
		0.0389, 0.0088, 0.2461, 0.0046, 0.0056, 0.1181, 0.1354, 0.0408, 0.0088, 
		0.2298, 0.0087, 0.0361, 0.0638, 0.0849, 0.1354], 
	4)


def test_annotate_seqlets_n_nearest(X, X_contrib):
	seqlets = recursive_seqlets(X_contrib)
	assert seqlets.shape == (24, 5)

	idxs, pvals = annotate_seqlets(X, seqlets, "tests/data/test.meme",
		n_nearest=3)
	
	assert idxs.shape == (24, 3)
	assert idxs.dtype == torch.int32

	assert pvals.shape == (24, 3)
	assert pvals.dtype == torch.float64


	idxs, pvals = annotate_seqlets(X, seqlets, "tests/data/test.meme",
		n_nearest=12)
	
	assert idxs.shape == (24, 12)
	assert idxs.dtype == torch.int32

	assert pvals.shape == (24, 12)
	assert pvals.dtype == torch.float64


def test_annotate_seqlets_reverse_complement(X, X_contrib):
	seqlets = recursive_seqlets(X_contrib)
	assert seqlets.shape == (24, 5)

	idxs0, pvals0 = annotate_seqlets(X, seqlets, "tests/data/test.meme",
		reverse_complement=True)
	idxs1, pvals1 = annotate_seqlets(X, seqlets, "tests/data/test.meme",
		reverse_complement=False)
	
	assert_raises(AssertionError, assert_array_almost_equal, idxs0, idxs1)
	assert_raises(AssertionError, assert_array_almost_equal, pvals0, pvals1)


###


def test_count_annotations_small():
	X = torch.Tensor([[0, 0], [0, 0], [0, 1], [1, 0]])
	y = count_annotations(X)

	assert_array_almost_equal(y, [[2, 1], [1, 0]])


def test_count_annotations(annotations):
	X = count_annotations(annotations)

	assert X.shape == (25, 7)
	assert X.sum() == 100

	assert_array_almost_equal(X.sum(dim=0), [16, 15, 14, 10, 18, 15, 12])
	assert_array_almost_equal(X.sum(dim=1), [6, 3, 2, 3, 4, 4, 3, 6, 5, 6, 4, 4, 
		3, 5, 2, 6, 2, 5, 4, 4, 3, 1, 3, 9, 3])


def test_count_annotations_dim0(annotations):
	X0 = count_annotations(annotations)
	X = count_annotations(annotations, dim=0)

	assert X.shape == (7,)
	assert X.sum() == 100
	assert_array_almost_equal(X, X0.sum(dim=0))


def test_count_annotations_dim1(annotations):
	X0 = count_annotations(annotations)
	X = count_annotations(annotations, dim=1)

	assert X.shape == (25,)
	assert X.sum() == 100
	assert_array_almost_equal(X, X0.sum(dim=1))


def test_count_annotations_dtype(annotations):
	X = count_annotations(annotations)
	assert X.dtype == torch.uint8

	X = count_annotations(annotations, dtype=torch.int64)
	assert X.dtype == torch.int64

	X = count_annotations(annotations, dtype=torch.float32)
	assert X.dtype == torch.float32


def test_count_annotations_shape(annotations):
	X = count_annotations(annotations)
	assert X.shape == (25, 7)

	X = count_annotations(annotations, shape=(30, 30))
	assert X.shape == (30, 30)

	assert_raises(RuntimeError, count_annotations, annotations, shape=(3, 3))
	assert_raises(RuntimeError, count_annotations, annotations, shape=(3, 10))
	assert_raises(RuntimeError, count_annotations, annotations, shape=(39, 3))


def test_count_annotations_raises(annotations):
	assert_raises(ValueError, count_annotations, annotations - 10)
	assert_raises(ValueError, count_annotations, annotations.T)
	assert_raises(ValueError, count_annotations, torch.randn(100, 2))
	assert_raises(ValueError, count_annotations, torch.cat([annotations, 
		annotations], dim=1))


###


def test_pairwise_annotations_small():
	X = torch.Tensor([[0, 0], [0, 0], [0, 1], [1, 0], [1, 1]]).type(torch.int8)
	y = pairwise_annotations(X)
	assert_array_almost_equal(y, [[0, 2], [2, 0]])

	y = pairwise_annotations(X, unique=False)
	assert_array_almost_equal(y, [[1, 3], [3, 0]])


def test_pairwise_annotations(annotations):
	y = pairwise_annotations(annotations)

	assert y.shape == (7, 7)
	assert y.dtype == torch.int64

	assert_array_almost_equal(y, [
		[0, 8, 7, 4, 6, 4, 7],
        [8, 0, 6, 4, 7, 6, 4],
        [7, 6, 0, 3, 7, 6, 3],
        [4, 4, 3, 0, 7, 3, 2],
        [6, 7, 7, 7, 0, 5, 3],
        [4, 6, 6, 3, 5, 0, 3],
        [7, 4, 3, 2, 3, 3, 0]
	])

	y = pairwise_annotations(annotations, unique=False)

	assert_array_almost_equal(y, [
		[ 4, 12, 10,  5,  8, 12, 12],
		[12,  3,  8,  5,  9, 15,  5],
		[10,  8,  2,  4,  9, 11,  4],
		[ 5,  5,  4,  1, 10,  6,  2],
		[ 8,  9,  9, 10,  4, 11,  3],
		[12, 15, 11,  6, 11, 6,  3],
		[12,  5,  4,  2,  3,  3,  2]
	])


def test_pairwise_annotations_raises(annotations):
	assert_raises(ValueError, pairwise_annotations, annotations - 10)
	assert_raises(ValueError, pairwise_annotations, annotations.T)
	assert_raises(ValueError, pairwise_annotations, torch.randn(100, 2))
	assert_raises(ValueError, pairwise_annotations, torch.cat([annotations, 
		annotations], dim=1))
	


# test_variant_effect.py
# Contact: Jacob Schreiber <jmschreiber91@gmail.com>

import numpy
import torch
import pytest
import pandas
import pyfaidx


from numpy.testing import assert_raises
from numpy.testing import assert_array_almost_equal

from .toy_models import FlattenDense
from .toy_models import SumModel
from .toy_models import SmallDeepSEA

from tangermeme.utils import random_one_hot

from tangermeme.variant_effect import substitution_effect
from tangermeme.variant_effect import deletion_effect
from tangermeme.variant_effect import insertion_effect


@pytest.fixture
def X():
	torch.manual_seed(0)
	return random_one_hot((3, 4, 100), random_state=0)


@pytest.fixture
def X_del():
	torch.manual_seed(0)
	return random_one_hot((3, 4, 103), random_state=0)


@pytest.fixture
def substitutions():
	return torch.tensor([
		[0, 4, 1],
		[0, 5, 2],
		[1, 2, 1],
		[1, 3, 1],
		[1, 8, 0]])


@pytest.fixture
def deletions():
	return torch.tensor([
		[0, 4],
		[0, 5],
		[1, 2],
		[1, 3],
		[1, 8]])


###


def test_substitution_effect(X, substitutions):
	model = SmallDeepSEA()
	y, y_var = substitution_effect(model, X, substitutions, device='cpu')

	assert y.shape == (3, 1)
	assert y_var.shape == (3, 1)

	assert_array_almost_equal(y, [[-0.0851], [-0.0855], [-0.0644]], 4)
	assert_array_almost_equal(y_var, [[-0.0835], [-0.0963], [-0.0644]], 4)


def test_substitution_effect_summodel(X, substitutions):
	model = SumModel()
	y, y_var = substitution_effect(model, X, substitutions, device='cpu')

	assert y.shape == (3, 4)
	assert y_var.shape == (3, 4)

	assert all(y.sum(dim=-1) == 100)
	assert all(y_var.sum(dim=-1) == 100)

	assert_array_almost_equal(y, [
		[25, 24, 19, 32],
        [26, 18, 29, 27],
        [28, 25, 21, 26]])
	assert_array_almost_equal(y_var, [
		[25, 25, 20, 30],
        [26, 20, 29, 25],
        [28, 25, 21, 26]])


###


def test_deletion_effect(X_del, deletions):
	model = SmallDeepSEA()
	y, y_var = deletion_effect(model, X_del, deletions, device='cpu')

	assert y.shape == (3, 1)
	assert y_var.shape == (3, 1)

	assert_array_almost_equal(y, [[-0.0851], [-0.0976], [-0.0940]], 4)
	assert_array_almost_equal(y_var, [[-0.0814], [-0.0852], [-0.0940]], 4)


def test_deletion_effect_summodel(X_del, substitutions):
	model = SumModel()
	y, y_var = deletion_effect(model, X_del, substitutions, device='cpu')

	assert y.shape == (3, 4)
	assert y_var.shape == (3, 4)

	assert all(y.sum(dim=-1) == 100)
	assert all(y_var.sum(dim=-1) == 100)

	assert_array_almost_equal(y, [
		[25, 24, 19, 32],
        [25, 18, 29, 28],
        [30, 25, 19, 26]])
	assert_array_almost_equal(y_var, [
		[25, 25, 20, 30],
        [23, 18, 30, 29],
        [30, 25, 19, 26]])

	assert_array_almost_equal(y, X_del[:, :, :100].sum(dim=-1))


###


def test_insertion_effect(X, substitutions):
	model = SmallDeepSEA()
	y, y_var = insertion_effect(model, X, substitutions, device='cpu')

	assert y.shape == (3, 1)
	assert y_var.shape == (3, 1)

	assert_array_almost_equal(y, [[-0.0851], [-0.0855], [-0.0644]], 4)
	assert_array_almost_equal(y_var, [[-0.0676], [-0.1054], [-0.0644]], 4)


def test_insertion_effect_summodel(X, substitutions):
	model = SumModel()
	y, y_var = insertion_effect(model, X, substitutions, device='cpu')

	assert y.shape == (3, 4)
	assert y_var.shape == (3, 4)

	assert_array_almost_equal(y, [
		[25, 24, 19, 32],
        [26, 18, 29, 27],
        [28, 25, 21, 26]])
	assert_array_almost_equal(y_var, [
		[24., 24., 20., 32.],
        [26., 19., 28., 27.],
        [28., 25., 21., 26.]])



# plot.py
# Contact: Jacob Schreiber <jmschreiber91@gmail.com>

import torch
import numpy
import pandas
import logomaker

from matplotlib import pyplot as plt
from matplotlib.colors import ListedColormap
import matplotlib

from .deep_lift_shap import deep_lift_shap


#plot_logo helper functions for placing annotations

def check_box_overlap(box1, box2):
	"""Check if annotation label text boxes overlap."""
	return not(box1.x0>=box2.x1 or box2.x0>=box1.x1 or box1.y0>=box2.y1 or box2.y0>=box1.y1)


def check_box_overlap_bar(box1, box2):
	"""Check if annotation bars overlap."""
	return not(box1.x0>=box2.x1 or box2.x0>=box1.x1 or box1.y0!=box2.y0)


def place_new_box(box, box_list, n_tracks=4, show_extra=True):
    """Place annotation text so that it does not overlap with existing text."""
    box_height = box.y1 - box.y0
    box.y0 -= box_height
    box.y1 -= box_height

    if len(box_list)==0:
        return box,0

    overlap_exists = any([check_box_overlap(box,box2) for box2 in box_list])
    steps_down_taken = 0
    while overlap_exists:
        steps_down_taken += 1
        if steps_down_taken == n_tracks: #beyond the n_tracks limit: make boxes smaller
                if not show_extra:
                        box.y0 -= box_height
                        box.y1 -= box_height
                        return box, steps_down_taken
                box.y1-= box_height
                box.y0-= box_height/2
                box_height = box_height/2
                overlap_exists = any([check_box_overlap(box,box2) for box2 in box_list])
                continue

        box.y0 -= box_height
        box.y1 -= box_height
        overlap_exists = any([check_box_overlap(box,box2) for box2 in box_list])

    return box, steps_down_taken


def place_new_bar(box, box_list, y_step=None, n_tracks=4, show_extra=True):
    """
    Find a position for a new annotation bar such that it does not overlap with previously plotted bars.
    """
    if y_step is None:
        raise ValueError("y_step must be provided.")

    if len(box_list)==0:
        return box, 0

    overlap_exists = any([check_box_overlap_bar(box,box2) for box2 in box_list])
    steps_down_taken = 0
    while overlap_exists:
        steps_down_taken += 1
        box.y0 -= y_step
        box.y1 -= y_step
        overlap_exists = any([check_box_overlap_bar(box,box2) for box2 in box_list])
    return box, steps_down_taken


def plot_logo(X_attr, ax=None, color=None, annotations=None, start=None, 
	end=None, ylim=None, spacing=4, n_tracks=4, score_key='score', 
	show_extra=True, show_score=True, annot_cmap="Set1"):
	"""Make a logo plot and optionally annotate it.

	This function will take in a matrix of weights for each character in a
	sequence and produce a plot where the characters have heights proportional
	to those weights. Attribution values from a predictive model are commonly
	used to weight the characters, but the weights can come from anywhere.

	Optionally, annotations can be provided in the form of a dataframe with
	contents described below in the parameters section. These annotations will
	be displayed underneath the characters in a manner that tries to avoid
	overlap across annotations.

	This function is largely a thin-wrapper around logomaker.


	Parameters
	----------
	X_attr: torch.tensor, shape=(4, -1)
			A tensor of the attributions. Can be either the hypothetical
			attributions, where the entire matrix has values, or the projected
			attributions, where only the actual bases have their attributions
			stored, i.e., 3 values per column are zero.

	ax: matplotlib.pyplot.subplot or None, optional
			The art board to draw on. If None, choose the current artboard.

	color: str or None, optional
			The color to plot all characters as. If None, plot according to
			standard coloring. Default is None.

	annotations: pandas.DataFrame, optional
			A set of annotations with the following columns in any order except for
			`motif_name`, which can be called anything but must come first:

					- motif_name: the name of the motif
					- start: the start of the hit relative to the window provided
					- end: the end of the hit relative to the window provided
					- strand: the strand the hit is on (optional)
					- score: the score of the hit

			These will probably come from the output of the hit caller. Default is
			None.

	start: int or None, optional
			The start of the sequence to visualize. Must be non-negative and cannot
			be longer than the length of `X_attr`. If None, visualize the full
			sequence. Default is None.

	end: int or None, optional
			The end of the sequence to visuaize. Must be non-negative and cannot be
			longer than the length of `X_attr`. If `start` is provided, `end` must 
			be larger. If None, visualize the full sequence. Default is None.

	ylim: tuple or None, optional
			The lower and upper bounds of the plot. Pass the bounds in here rather
			than setting them after calling this function if you want the annotation
			spacing to adjust to it. If None, use the default bounds. Default is
			None.

	n_tracks: int, optional
			The number of rows of annotation labels to plot with bars before simply
			putting the name of the motif. Default is 4.

	score_key: str, optional
			When annotations are provided, the name of the key to use as a score.
			Must have the semantics that a higher value means a "better" annotation.
			Default is 'score'.

	show_extra: bool, optional
			Whether to show motif names past the `n_tracks` number of rows. 
			If False, do not show those motifs. Default is True.

	show_score: bool, optional
			Whether to show the score of the hit. Sometimes, the annotation can take up
			too much space already and the scoring information is not as helpful, so
			disabling this will only display the motif hit name. Default is True.

	annot_cmap: str, list, or matplotlib.colors.ListedColormap, optional
			The colormap to use for the annotations. Rows of annotation labels receive 
			distinct colors, bars are colored according to their corresponding label.
			If a string, must be a valid matplotlib qualitative colormap name 
			If a list, must be a list of colors (list of colornames or list of RGB tuples).
			Labels are colorcoded if there is more than 1 row of annotations, otherwise black.


	Returns
	-------
	ax: plt.subplot
			A subplot that contains the plot.
	"""

	try:
			import matplotlib.pyplot as plt
	except:
			raise ImportError("Must install matplotlib before using.")

	###
	# Main glyph plotting code
	###


	if start is not None and end is not None:
		X_attr = X_attr[:, start:end]

	if ax is None:
		ax = plt.gca()

	df = pandas.DataFrame(X_attr.T, columns=['A', 'C', 'G', 'T'])
	df.index.name = 'pos'

	logo = logomaker.Logo(df, ax=ax)
	logo.style_spines(visible=False)

	if color is not None:
		alpha = numpy.array(['A', 'C', 'G', 'T'])
		seq = ''.join(alpha[numpy.abs(df.values).argmax(axis=1)])
		logo.style_glyphs_in_sequence(sequence=seq, color=color)


	###
	# Handling potentially overlapping annotations
	###


	#Set annotation colormap
	if type(annot_cmap) == str:
		cmap = plt.get_cmap(annot_cmap)
	elif type(annot_cmap) == list:
		cmap = ListedColormap(annot_cmap)
	elif type(annot_cmap) == matplotlib.colors.ListedColormap:
		cmap = annot_cmap

	if annotations is not None:
		start, end = start or 0, end or X_attr.shape[-1]

		annotations_ = annotations[annotations['start'] > start]
		annotations_ = annotations_[annotations_['end'] < end]
		annotations_ = annotations_.sort_values(["start"], ascending=True)

		if len(annotations_) > 0:
			ylim = ylim or max(abs(X_attr.min()), abs(X_attr.max()))
			ax.set_ylim(-ylim, ylim)
			r = ylim*2
			y_offset_bars=ax.get_ylim()[0]/8
			y_offset_labels = 0

			#deterrmine label text size and line width according to figure size
			bbox = ax.get_position()
			fig_width, fig_height = ax.get_figure().get_size_inches()
			width_in = bbox.width * fig_width
			height_in = bbox.height * fig_height
			labelsize=width_in*1.1
			linewidth = width_in*0.25

			#plotting annotation labels
			label_box_objects = []
			visible_label_box_objects = []
			label_text_boxes = []
			text_box_colors = []
			
			for i,(_, row) in enumerate(annotations_.iterrows()):
				motif = row.values[0]
				motif_start = int(row['start'])
				motif_end = int(row['end'])
				score = row[score_key]
				motif_start -= start
				motif_end -= start

				#define label text
				text = str(motif)
				if show_score:
						text += ' ({:3.3})'.format(score)
				text=text.replace(" ", "\n")

				#plot text box in top most position...
				text_box = ax.text(motif_start, y_offset_labels, text, fontsize=labelsize)
				ax.get_figure().canvas.draw()
				bbox = text_box.get_window_extent()

				#...shift box down if it overlaps with previously drawn boxes
				bbox_new, steps_down_taken = place_new_box(bbox, label_text_boxes, n_tracks=n_tracks,show_extra=show_extra)
				bbox_new_transformed = bbox_new.transformed(ax.transData.inverted())

				#color the text according to the number of downshifts
				text_color = cmap(steps_down_taken)
				if steps_down_taken >= n_tracks: #if box is beyond the n_tracks limit, plot in smaller font size and in light grey.
						text_color = (0.7,0.7,0.7)
						text_box.set_fontsize(labelsize/2)
						if not show_extra:
								text_color = (1,1,1,0)
				else:
						visible_label_box_objects.append(text_box)

				text_box.set_position((bbox_new_transformed.x0, bbox_new_transformed.y0))
				text_box.set_color(text_color)
				text_box_colors.append(text_color)
				label_text_boxes.append(bbox_new)
				label_box_objects.append(text_box)

			#plotting annotation bars
			bars_box_objects = []
			bars_boxes = []
			bars_ymins=[]
			for i,(_, row) in enumerate(annotations_.iterrows()):
				motif = row.values[0]
				motif_start = int(row['start'])
				motif_end = int(row['end'])
				score = row[score_key]
				motif_start -= start
				motif_end -= start

				bar_color = text_box_colors[i]
				if bar_color == (0.7,0.7,0.7) or bar_color == (1,1,1,0): #for labels beyond the n_tracks limit, no bar is plotted.
						continue

				xp = [motif_start-0.1, motif_end-0.5]
				yp = [y_offset_bars, y_offset_bars]

				#plot bar in topmost place...
				bar = ax.plot(xp, yp, color='0.3', linewidth=linewidth)
				ax.get_figure().canvas.draw()

				#...shift bar down if it overlaps with previously drawn bars
				bar_box = bar[0].get_window_extent()
				bar_box_new, steps_down_taken = place_new_bar(bar_box, bars_boxes, y_step=linewidth*2, n_tracks=n_tracks,show_extra=show_extra)
				bar_box_new_transformed = bar_box_new.transformed(ax.transData.inverted())
				bar[0].set_ydata([bar_box_new_transformed.y0, bar_box_new_transformed.y1])
				bars_boxes.append(bar_box_new)
				bar[0].set_color(bar_color)
				bars_ymins.append(bar_box_new_transformed.y0)
				bars_box_objects.append(bar[0])


			#shift text boxes down under the lowest bar
			bars_ymin = min(bars_ymins)
			for label_box in label_box_objects:
					label_box.set_y(label_box.get_position()[1] + bars_ymin) 

			#if there is only one row of annotations, set colors to black
			if len(set([vis_box.get_color() for vis_box in visible_label_box_objects])) == 1:
					for label_box in visible_label_box_objects:
							label_box.set_color((0,0,0))
					for bar_box in bars_box_objects:
							bar_box.set_color((0,0,0))

	return logo


def plot_categorical_scatter(X, colors=None, **kwargs):
	"""A scatterplot of category weights across a seq, useful for attributions.
	
	Frequently, when you calculate attributions you are considering a sequence
	that is too long to be easily visualized in the standard character format,
	e.g. in `plot_logo`. One could scan the sequence a few times to find the
	regions of high attribution but that is also time-consuming. This function
	creates a scatterplot where the coloring corresponds to the category (e.g.,
	which nucleotide is there) and the value is the weight encoded in the matrix.
	Basically, this is a way to consider attributions on an entire sequence and
	get a sense for which areas you may want to follow-up with.
	
	This function is not meant solely for attributions, though. You can put any
	value in the matrix to be visualized.
	
	
	Parameters
	----------
	X: torch.Tensor, shape=(alphabet_len, length)
		A sequence to be visualized.
	
	colors: list or None, optional
		The colors to use for each category. By default, uses the standard
		nucleotide color scheme.
	
	**kwargs: any additional arguments
	"""
	
	if colors is None:
		c = [[0, 0.5, 0], [0, 0, 1], [1, 0.65, 0], [1, 0, 0]]

	pos = numpy.arange(X.shape[-1])
	for i in range(X.shape[0]):
		idxs = torch.abs(X).argmax(axis=0) == i    
		plt.scatter(pos[idxs], X.sum(axis=0)[idxs], c=[c[i]], **kwargs)


def plot_attributions(models, X, func=deep_lift_shap, attribute_kwargs=None, 
	plot_kwargs=None, layout=None):
	"""A convenience function for calculating and then plotting attributions.
	
	This function will use one or more models and calculate attributions on one or
	more sequences. These attributions are then plotted according to the format in
	layout, and returned to the user. This is just a wrapped around using an
	attribution function and `plot_logo`, and so additional arguments can be 
	passed into the attribution or the plotting functions using the respective
	kwargs arguments. `start` and `end` are likely 
	to be common arguments to pass
	into `plot_kwargs`.
	
	There are three signatures supported. (1) A list of models are provided with
	the same length as the first dimension in `X`. The i-th model gets applied to
	the i-th element of X. This is the most flexible, as the same model or
	sequence could be passed in multiple times. (2) Rather than a list of models,
	a single model object is provided, and gets applied to all the elements of X.
	(3) Rather than multiple sequences, a single sequence is provided for X as a
	2D tensor. in which case all models are applied to it.
	
	By default, `deep_lift_shap` is used to calculate attributions but any
	function can be used as long as it has the same signature.


	Parameters
	----------
	models: torch.nn.Module or list of torch.nn.Modules
		One or a set of models to apply to the sequence/s.
	
	X: torch.Tensor, shape=(-1, len(alphabet), -1) or (len(alphabet), -1)
		One or more sequences 

	func: func, optional
		The attribution function to use. Default is deep_lift_shap.
	
	attribute_kwargs: dict or None, optional
		Arguments to pass into `func`. Default is None.
	
	plot_kwargs: dict or None, optional
		Arguments to pass into `func`. Default is None.

	layout: 2-tuple or None, optional
		A layout of the subplots, the first two numbers passed into
		`plt.subplot`. If None, assume (n, 1).

	Returns
	-------
	axs: list of plt.axis
		A list of artboards.
	
	attributions: torch.Tensor
		The attributions being visualized here. This is the full output from `func`,
		even when only a portion is visualized.
	"""

	try:
		import matplotlib.pyplot as plt
	except:
		raise ImportError("Must install matplotlib before using.")
	
	
	if isinstance(models, torch.nn.Module):
		if X.ndim != 3:
			raise ValueError("X must have 3 dimensions when only providing one model.")
		
		models = [models for _ in range(X.shape[0])]
		
	elif X.ndim == 2:
		X = X.repeat(len(models), 1, 1)
	
	else:
		if len(models) != len(X):
			raise ValueError("X must have first dimension equal to number of models.")
	
	if layout is None:
		layout = (len(X), 1)
	else:
		if layout[0] * layout[1] < len(X):
			raise ValueError("Layout must have >= number of models.")
			
	attribute_kwargs = attribute_kwargs or {}
	plot_kwargs = plot_kwargs or {}
	
	X_attrs = []
	for i, model in enumerate(models):
		X_attr = deep_lift_shap(model, X[i:i+1], **attribute_kwargs)
		X_attrs.append(X_attr)
	X_attrs = torch.cat(X_attrs, dim=0).detach()
	
	axs = []
	for i in range(len(X_attrs)):
		ax = plt.subplot(*layout, i+1)
		axs.append(ax)
		
		plot_logo(X_attrs[i], ax=ax, **plot_kwargs)
	
	return axs, X_attrs

def plot_pwm(pwm, name=None, alphabet=['A', 'C', 'G', 'T'], eps=1e-7):
	"""Plots an information-content weighted PWM and its reverse complement.

	This function takes in a PWM, where the sum across all values in the
	alphabet is equal to 1, and plots the information-content weighted version
	of it, as well as the reverse complement. This should be used when you want
	to visualize a motif, perhaps from a motif database.


	Parameters
	----------
	pwm: torch.Tensor or numpy.ndarray, shape=(len(alphabet), length)
		The PWM to visualize. The rows must sum to 1.

	name: str or None, optional
		The name to put as the title for the plots. If None, do not put
		anything. Default is None.

	alphabet: list, optional
		A list of characters that comprise the alphabet. Default is
		['A', 'C', 'G', 'T'].

	eps: float, optional
		A small pseudocount to add to counts to make the log work correctly.
		Default is 1e-7.
	"""

	if isinstance(pwm, torch.Tensor):
		pwm = pwm.numpy(force=True)
	
	bg = 0.25 * numpy.log(0.25) / numpy.log(2)

	
	plt.figure(figsize=(8, 2.5))
	ax = plt.subplot(121)
	plt.title(name)
	
	ic = pwm * numpy.log(pwm + eps) / numpy.log(2) - bg
	ic = numpy.sum(ic, axis=0, keepdims=True)
	plot_logo(pwm * ic, ax=ax)
	plt.xlabel("Motif Position")
	plt.ylabel("Information Content (Bits)", fontsize=10)
	
	
	ax = plt.subplot(122)
	plt.title(name + "RC" if name is not None else "RC")
	pwm = pwm[::-1, ::-1]
	ic = pwm * numpy.log(pwm + eps) / numpy.log(2) - bg
	ic = numpy.sum(ic, axis=0, keepdims=True)
	plot_logo(pwm * ic, ax=ax)
	plt.xlabel("Motif Position")
	plt.ylabel("Information Content (Bits)", fontsize=10)
	
	plt.tight_layout()
	plt.show()
	


# annotate.py
# Author: Jacob Schreiber <jmschreiber91@gmail.com>

import numpy
import torch
import pandas

from .io import read_meme
from .utils import _validate_input

from memelite import tomtom


def annotate_seqlets(X, seqlets, motifs, n_nearest=1, n_jobs=-1, **kwargs):
	"""Annotate a set of seqlets according to a motif database using TOMTOM.

	This function takes in a set of seqlets and a motif database and assigns
	to each seqlet the most significant motif match. This match is done using
	TOMTOM on just the one-hot encoded sequence for the seqlet and the PWM
	for the motifs. If no motifs have a statistical significance below the
	provided threshold, an index of -1 is returned.

	The computation is independent for each seqlet, so ordering of seqlets
	should not matter, not should running this function on a subset of seqlets
	versus the full set.


	Parameters
	----------
	X: torch.Tensor, shape=(n_examples, len(alphabet), length)
		A one-hot encoded set of sequences. These are only relevant for
		extracting the sequences spanned by each seqlet.

	seqlets: pandas.DataFrame, shape=(n_seqlets, 3+)
		A BED-formatted set of seqlets to annotate. The first three columns
		must correspond to the example index in `X`, the start, and the end
		(both base-0). The annotations can have additional columns, but those
		are ignored.

	motifs: dict or str
		A dictionary of motifs where the keys are motif names and the values
		are PWMs with shape (len(alphabet), len(motif)), or a string to a
		MEME-formatted file.

	n_nearest: int, optional
		The number of motif matches to return for each seqlet, starting with
		the most significant hit. Default is 1.

	n_jobs: int, optional
		The number of threads to run TOMTOM on in parallel. -1 means use all
		available threads. Default is -1.

	**kwargs: arguments, optional
		Any other arguments to pass into the TOMTOM algorithm.


	Returns
	-------
	idxs: torch.Tensor, shape=(len(seqlets), n_nearest)
		The index of the most significant motif hit(s) for each seqlet.

	pvals: torch.Tensor, shape=(len(seqlets), n_nearest)
		The p-value of the returned significant hits.
	"""

	if isinstance(motifs, str):
		motifs = read_meme(motifs)

	motif_pwms = list(motifs.values())

	X_seqlets = []
	for example_idx, start, end in seqlets.iloc[:, :3].values:
		X_seqlets.append(X[example_idx, :, start:end])

	p_values, _, _, _, _, idxs = tomtom(X_seqlets, motif_pwms, 
		n_nearest=n_nearest, n_jobs=n_jobs, **kwargs)

	return torch.from_numpy(idxs).type(torch.int32), torch.from_numpy(p_values)


def count_annotations(X, dtype=torch.uint8, shape=None, dim=None):
	"""A method for counting the annotations for each example.

	This function takes in a tensor of (example_idx, annotation_idx) pairs and
	returns a tensor of counts where each row is an example, each column is an
	annotation, and the value is the number of times that annotation appears
	in that example. The dimensions of the returned matrix will then be the
	maximum example_idx and motif_idx value even if intermediary values are
	not observed.


	Parameters
	----------
	X: torch.Tensor, shape=(n_annotations, 2) or tuple of two vectors
		A tensor of annotations where the first column is the example_idx and
		the second column is the annotation_idx. Both should be integers. 

	dtype: torch.dtype, optional
		The dtype of the returned matrix. Default is torch.uint8.

	shape: tuple or None, optional
		A user-defined shape of the returned count matrix. Use this if you are
		not sure whether all examples or annotations are observed in `X` but you
		want to have a constant shape. If None, derive shape from the maximum
		values in `X`. Default is None.

	dim: None, 0, or 1, optional
		Whether to aggregate the counts along one of the axes. If set to None,
		the full count matrix will be returned. If set to 0 the first dimension
		is summed over, returning the number of occurences of each annotation. 
		If set to 1, the second dimension is summed over, returning the number
		of annotations per example. Default is None. 


	Returns
	-------
	y: torch.Tensor, shape=(max(example_idx), max(motif_idx))
		A sparse tensor where each row is an example and each column is an
		annotation and the values within are the number of times each annotation
		appears in each example.
	"""

	if isinstance(X, (list, tuple)):
		x_ = []
		for x in X:
			if isinstance(x, pandas.Series):
				x = x.values

			if isinstance(x, numpy.ndarray):
				x = torch.from_numpy(x)

			x_.append(x)

		X = torch.vstack(x_).T

	_validate_input(X, 'X', shape=(-1, 2), min_value=0)

	X = X.type(torch.int64)
	X_ones = torch.ones(len(X), dtype=dtype)

	n_examples, n_annotations = X.max(dim=0).values + 1

	if shape is not None:
		if n_examples > shape[0] or n_annotations > shape[1]:
			raise RuntimeError("Observed maximum indices {} but".format(
				(n_examples, n_annotations)) + " provided shape {}".format(
				shape))
		
		n_examples, n_annotations = shape


	if dim == 0:
		y = torch.zeros(n_annotations, dtype=dtype)
		y.scatter_add_(0, X[:, 1], X_ones)
	
	elif dim == 1:
		y = torch.zeros(n_examples, dtype=dtype)
		y.scatter_add_(0, X[:, 0], X_ones)
	
	else:
		X_idxs = X[:, 0] * n_annotations + X[:, 1]
		
		y = torch.zeros(n_examples * n_annotations, dtype=dtype)
		y.scatter_add_(0, X_idxs, X_ones)
		y = y.reshape(n_examples, n_annotations)
	
	return y


def pairwise_annotations(X, dtype=torch.int64, unique=True, symmetric=True, shape=None):
	"""Returns the number of times pairs of annotations occur in an example.

	This function takes in a tensor of (example_idx, annotation_idx) pairs and
	returns a tensor of counts where each row is an annotation_idx and each
	column is also an annotation_idx, and the values within are the number of
	times that the each pair of annotations appears in the same example.

	The returned matrix will be a symmetric matrix that has a total number of
	counts equal to twice the number of provided examples.


	Parameters
	----------
	X: torch.Tensor, shape=(n_annotations, 2)
		A tensor of annotations where the first column is the example_idx and
		the second column is the annotation_idx. Both should be integers. 

	dtype: torch.dtype, optional
		The dtype of the returned matrix. Default is torch.int64.

	unique: bool, optional
		Whether to count only unique pairs within each example or each instance.
		For example, if set to True, an example with 3 instances of KL4 would
		contribute only a single KLF4-KLF4 interaction to the matrix, whereas
		if set to True, would contribute many. Default is True.

	symmetric: bool, optional
		Whether to return a symmetric matrix or one where the row is the first
		element in `X` and the column is the subsequent element in `X`. If
		symmetric, the diagonal is NOT double counted. Default is True.

	shape: int or None, optional
		The number of rows and columns to use in the matrix. If None, infer the
		number from `X`. Default is None.


	Returns
	-------
	y: torch.Tensor, shape=(max(example_idx), max(motif_idx))
		A sparse tensor where each row is an example and each column is an
		annotation and the values within are the number of times each annotation
		appears in each example.
	"""

	if isinstance(X, (list, tuple)):
		x_ = []
		for x in X:
			if isinstance(x, pandas.Series):
				x = x.values

			if isinstance(x, numpy.ndarray):
				x = torch.from_numpy(x)

			x_.append(x)

		X = torch.vstack(x_).T

	_validate_input(X, 'X', shape=(-1, 2), min_value=0)

	n_examples, n_annotations = X.max(dim=0).values + 1

	if shape is not None:
		if n_annotations > shape:
			raise RuntimeError("Observed maximum indices {} but".format(
				(n_examples, n_annotations)) + " provided shape {}".format(
				shape))
		
		n_annotations = shape

	example_annotations = [[] for i in range(n_examples)]
	for example_idx, annotation_idx in X:
		example_annotations[example_idx].append(annotation_idx)

	y = torch.zeros(n_annotations, n_annotations, dtype=dtype).numpy()
	for annotations in example_annotations:
		if unique:
			annotations = numpy.unique(annotations)
		
		for i, idx0 in enumerate(annotations[:-1]):
			for j, idx1 in enumerate(annotations[i+1:]):
				y[idx0, idx1] += 1

				if symmetric and idx0 != idx1:
					y[idx1, idx0] += 1

	return torch.from_numpy(y)


def pairwise_annotations_spacing(X, max_distance=100, dtype=torch.uint8, 
	symmetric=True, shape=None):
	"""Finds the number of times each annotation pairs happens at each distance.

	This function takes in a tensor of (example_idx, annotation_idx, start, end) 
	tuples and returns a tensor of counts where the first two dimensions are 
	annotation indexes and the third dimension is the spacing between the pair. 
	The values within the tensor are the count of the number of times that pair 
	of annotations is found with that spacing. Importantly, distance is 
	calculated between the end of the motif on the left and the start of the 
	motif on the right. Put another way, it is the number of nucleotides between 
	each motif and so is invariant to their lengths.

	The returned tensor will be extremely sparse, so be careful about setting
	the maximum distance to a value that is very high. A sparse tensor could
	be used here, but given how cheap memory is getting this function is meant
	to be a quick solution that handles most cases.


	Parameters
	----------
	X: torch.Tensor, shape=(n_annotations, 4)
		A tensor of annotations where the columns should be the example_idx,
		annotation_idx, start of the annotation, and end of the annotation.
		The end of the annotation should not be inclusive, so two adjacent
		annotations with zero spacing between them should have the same integer
		index for the end of one motif and the start of the next.

	max_distance: int, optional
		The maximum distance between two annotations in the same example to
		consider. Default is 100.

	dtype: torch.dtype, optional
		The dtype of the returned matrix. Default is torch.uint8.

	symmetric: bool, optional
		Whether to return a symmetric matrix or one where the row is the first
		element in `X` and the column is the subsequent element in `X`. If
		symmetric, the diagonal is NOT double counted. Default is True.

	shape: int or None, optional
		The number of rows and columns to use in the matrix. If None, infer the
		number from `X`. Default is None.


	Returns
	-------
	y: torch.Tensor, shape=(max(example_idx), max(motif_idx))
		A sparse tensor where each row is an example and each column is an
		annotation and the values within are the number of times each annotation
		appears in each example.
	"""

	if isinstance(X, (list, tuple)):
		x_ = []
		for x in X:
			if isinstance(x, pandas.DataFrame):
				x = torch.from_numpy(x.values)

			elif isinstance(x, numpy.ndarray):
				x = torch.from_numpy(x)

			if x.ndim == 1:
				x = x.unsqueeze(1)

			x_.append(x)

		X = torch.cat(x_, axis=1)[:, [0, 3, 1, 2]]

	elif isinstance(X, pandas.DataFrame):
		X = torch.from_numpy(X.values)

	_validate_input(X, 'X', shape=(-1, 4), min_value=0)

	n_examples, n_annotations = X.max(dim=0).values[:2] + 1

	if shape is not None:
		if n_annotations > shape:
			raise RuntimeError("Observed maximum indices {} but".format(
				(n_examples, n_annotations)) + " provided shape {}".format(
				shape))
		
		n_annotations = shape


	example_annotations = [[] for i in range(n_examples)]

	y = torch.zeros(n_annotations, n_annotations, max_distance, 
		dtype=dtype).numpy()

	for example_idx, annotation_idx, start, end in X:
		example_annotations[example_idx].append((annotation_idx, start, end))

	for annotations in example_annotations:
		for i, (idx0, start0, end0) in enumerate(annotations[:-1]):
			for j, (idx1, start1, end1) in enumerate(annotations[i+1:]):
				if start0 < start1:
					d = start1 - end0
					if d > max_distance:
						continue

					y[idx0, idx1, d] += 1
					if symmetric and idx0 != idx1:
						y[idx1, idx0, d] += 1

				else:
					d = start0 - end1
					if d > max_distance:
						continue

					y[idx1, idx0, d] += 1
					if symmetric and idx0 != idx1:
						y[idx0, idx1, d] += 1 

	return torch.from_numpy(y)



# test_design.py
# Contact: Jacob Schreiber <jmschreiber91@gmail.com>

import torch
import pytest

from tangermeme.utils import random_one_hot
from tangermeme.ersatz import substitute

from tangermeme.predict import predict

from tangermeme.design import screen
from tangermeme.design import greedy_substitution
from tangermeme.design import greedy_marginalize

from tangermeme.io import read_meme
from tangermeme.utils import characters

from .toy_models import SumModel
from .toy_models import SmallDeepSEA

from numpy.testing import assert_raises
from numpy.testing import assert_array_almost_equal


torch.manual_seed(0)
torch.use_deterministic_algorithms(True, warn_only=True)


@pytest.fixture
def X():
	return random_one_hot((1, 4, 100), random_state=0)


@pytest.fixture
def X_marg():
	return random_one_hot((5, 4, 100), random_state=0)


@pytest.fixture
def motifs():
	return [characters(m, force=True) for m in 
		read_meme("tests/data/test.meme").values()]


###


def test_screen():
	torch.manual_seed(0)
	model = SmallDeepSEA()
	y = [10]

	X_hat = screen(model, (4, 100), y, device='cpu', max_iter=1, random_state=0)

	assert X_hat.shape == (1, 4, 100)
	assert X_hat.dtype == torch.int8
	assert X_hat.sum(dim=1).min() == 1
	assert all(torch.unique(X_hat) == torch.tensor([0, 1], dtype=torch.int8))

	assert_array_almost_equal(X_hat[:, :, :20], [
		[[1, 0, 0, 0, 1, 0, 0, 0, 0, 1, 0, 1, 0, 1, 0, 0, 1, 0, 1, 0],
         [0, 0, 0, 0, 0, 1, 0, 1, 1, 0, 1, 0, 0, 0, 0, 1, 0, 0, 0, 1],
         [0, 0, 0, 1, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 1, 0, 0],
         [0, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0]]
	])

	y_hat = predict(model, X_hat, device='cpu')

	assert_array_almost_equal(y_hat, [[-0.054611]])

	##

	X_hat = screen(model, (4, 100), y, device='cpu', max_iter=10, random_state=0)

	assert X_hat.shape == (1, 4, 100)
	assert X_hat.dtype == torch.int8
	assert all(torch.unique(X_hat) == torch.tensor([0, 1], dtype=torch.int8))

	assert_array_almost_equal(X_hat[:, :, :20], [
		[[0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 1, 1, 0, 0, 0, 0, 0, 0],
         [0, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0],
         [0, 0, 0, 0, 0, 1, 0, 1, 0, 1, 0, 1, 0, 0, 0, 1, 1, 0, 1, 0],
         [1, 0, 0, 0, 1, 0, 0, 0, 1, 0, 1, 0, 0, 0, 0, 0, 0, 1, 0, 1]]
	])

	y_hat = predict(model, X_hat, device='cpu')

	assert_array_almost_equal(y_hat, [[-0.03848]])


def test_screen_summodel():
	torch.manual_seed(0)
	model = SumModel()
	y = [10, 0, 0, 0]

	X_hat = screen(model, (4, 10), y, device='cpu', max_iter=25, random_state=0)

	assert X_hat.shape == (1, 4, 10)
	assert X_hat.dtype == torch.int8
	assert all(torch.unique(X_hat) == torch.tensor([0, 1], dtype=torch.int8))

	assert_array_almost_equal(X_hat[:, :, :20], [
		[[0, 1, 1, 1, 1, 1, 1, 0, 0, 1],
         [0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
         [1, 0, 0, 0, 0, 0, 0, 0, 0, 0],
         [0, 0, 0, 0, 0, 0, 0, 1, 1, 0]]
	])

	y_hat = predict(model, X_hat, device='cpu')

	assert_array_almost_equal(y_hat, [[7., 0., 1., 2.]])


def test_screen_n_best():
	torch.manual_seed(0)
	model = SmallDeepSEA()
	y = [10]
	
	X_hat = screen(model, (4, 100), y, device='cpu', n_best=7, 
		max_iter=10, random_state=0)

	assert X_hat.shape == (7, 4, 100)
	assert X_hat.dtype == torch.int8
	assert all(torch.unique(X_hat) == torch.tensor([0, 1], dtype=torch.int8))

	assert_array_almost_equal(X_hat[:3, :, :10], [
		[[0, 0, 0, 0, 0, 0, 1, 0, 0, 0],
         [0, 1, 1, 1, 0, 0, 0, 0, 0, 0],
         [0, 0, 0, 0, 0, 1, 0, 1, 0, 1],
         [1, 0, 0, 0, 1, 0, 0, 0, 1, 0]],

        [[0, 0, 0, 1, 0, 0, 0, 0, 0, 0],
         [1, 0, 0, 0, 0, 0, 0, 1, 0, 0],
         [0, 1, 0, 0, 0, 0, 1, 0, 0, 0],
         [0, 0, 1, 0, 1, 1, 0, 0, 1, 1]],

        [[1, 0, 0, 0, 0, 0, 1, 0, 0, 1],
         [0, 1, 0, 0, 0, 0, 0, 0, 0, 0],
         [0, 0, 0, 0, 0, 1, 0, 1, 0, 0],
         [0, 0, 1, 1, 1, 0, 0, 0, 1, 0]]])

	y_hat = predict(model, X_hat, device='cpu')

	assert_array_almost_equal(y_hat, [
		[-0.0385],
        [-0.0439],
        [-0.0454],
        [-0.0457],
        [-0.0466],
        [-0.0496],
        [-0.0510]
	], 4)


def test_screen_no_y():
	torch.manual_seed(0)
	model = SmallDeepSEA()

	X_hat = screen(model, (4, 100), device='cpu', random_state=0, max_iter=15)
	y_hat = predict(model, X_hat, device='cpu')

	assert_array_almost_equal(y_hat, [[-0.085115]])


def test_screen_custom_func():
	torch.manual_seed(0)
	def onlyAs(shape, random_state):
		return random_one_hot(shape, probs=[1.0, 0.0, 0.0, 0.0])
	
	model = SmallDeepSEA()
	y = [10]
	
	X_hat = screen(model, (4, 100), y, device='cpu', max_iter=10, 
		random_state=0, func=onlyAs)

	assert X_hat.shape == (1, 4, 100)
	assert X_hat.dtype == torch.int8
	assert all(torch.unique(X_hat) == torch.tensor([0, 1], dtype=torch.int8))

	assert_array_almost_equal(X_hat[:, :, :20], [
		[[1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1],
         [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
         [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
         [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0]]
	])

	y_hat = predict(model, X_hat, device='cpu')

	assert_array_almost_equal(y_hat, [[-0.0841]], 4)


def test_screen_custom_kwargs():
	torch.manual_seed(0)
	model = SmallDeepSEA()
	y = [10]
	
	X_hat = screen(model, (4, 100), y, device='cpu', max_iter=1, 
		additional_func_kwargs={'probs': [1, 0, 0, 0]})

	assert X_hat.shape == (1, 4, 100)
	assert X_hat.dtype == torch.int8
	assert all(torch.unique(X_hat) == torch.tensor([0, 1], dtype=torch.int8))

	assert_array_almost_equal(X_hat[:, :, :20], [
		[[1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1],
         [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
         [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
         [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0]]
	])

	y_hat = predict(model, X_hat, device='cpu')

	assert_array_almost_equal(y_hat, [[-0.0841]], 4)


def test_screen_raises():
	torch.manual_seed(0)
	model = SmallDeepSEA()
	y = [[0.5]]
	
	assert_raises(RuntimeError, screen, model, (5, 100), y, device='cpu')
	assert_raises(ValueError, screen, model, (1, 4, 100), y, device='cpu')
	assert_raises(ValueError, screen, model, (100,), y, device='cpu')


###


def test_greedy_substitution(X, motifs):
	torch.manual_seed(0)
	model = SmallDeepSEA()
	y = [[10]]

	X_hat = greedy_substitution(model, X, y, motifs, device='cpu', max_iter=1)

	assert X_hat.shape == (1, 4, 100)
	assert X_hat.dtype == torch.int8
	assert X_hat.sum(dim=1).min() == 1
	assert all(torch.unique(X_hat) == torch.tensor([0, 1], dtype=torch.int8))

	y_hat = predict(model, X_hat, device='cpu')

	assert_array_almost_equal(y_hat, [[-0.058557]])

	##

	X_hat = greedy_substitution(model, X, y, motifs, device='cpu', max_iter=4)

	assert X_hat.shape == (1, 4, 100)
	assert X_hat.dtype == torch.int8
	assert X_hat.sum(dim=1).min() == 1
	assert all(torch.unique(X_hat) == torch.tensor([0, 1], dtype=torch.int8))
	
	y_hat = predict(model, X_hat, device='cpu')

	assert_array_almost_equal(y_hat, [[-0.018134]])


def test_greedy_nucleotide_substitution(X):
	torch.manual_seed(0)
	model = SmallDeepSEA()
	y = [[10]]

	X_hat = greedy_substitution(model, X, y, device='cpu', max_iter=1)

	assert X_hat.shape == (1, 4, 100)
	assert X_hat.dtype == torch.int8
	assert X_hat.sum(dim=1).min() == 1
	assert all(torch.unique(X_hat) == torch.tensor([0, 1], dtype=torch.int8))
	assert abs(X_hat - X).sum() == 2

	y_hat = predict(model, X_hat, device='cpu')

	assert_array_almost_equal(y_hat, [[-0.076893]])

	##

	X_hat = greedy_substitution(model, X, y, device='cpu', max_iter=4)

	assert X_hat.shape == (1, 4, 100)
	assert X_hat.dtype == torch.int8
	assert X_hat.sum(dim=1).min() == 1
	assert all(torch.unique(X_hat) == torch.tensor([0, 1], dtype=torch.int8))
	assert abs(X_hat - X).sum() == 8
	
	y_hat = predict(model, X_hat, device='cpu')

	assert_array_almost_equal(y_hat, [[-0.057514]])


def test_greedy_nucleotide_substitution_only_As(X):
	torch.manual_seed(0)
	model = SmallDeepSEA()
	motifs = ['A']
	y = [[10]]

	X_hat = greedy_substitution(model, X, y, motifs, device='cpu', max_iter=1, 
		reverse_complement=False)
	assert X_hat.sum(dim=(0, -1))[0] == X.sum(dim=(0, -1))[0] + 1

	X_hat = greedy_substitution(model, X, y, motifs, device='cpu', max_iter=4, 
		reverse_complement=False)
	assert X_hat.sum(dim=(0, -1))[0] == X.sum(dim=(0, -1))[0] + 4


def test_greedy_substitution_input_mask(X):
	torch.manual_seed(0)
	input_mask = torch.zeros(100, dtype=bool)
	input_mask[54:58] = True

	model = SmallDeepSEA()
	y = [[10]]

	X_hat = greedy_substitution(model, X, y, device='cpu', input_mask=input_mask,
		max_iter=4)

	assert_array_almost_equal(X[:, :, :54], X_hat[:, :, :54])
	assert_array_almost_equal(X[:, :, 58:], X_hat[:, :, 58:])
	assert abs(X[:, :, 54:58] - X_hat[:, :, 54:58]).sum() == 6


def test_greedy_substitution_output_mask(X):
	torch.manual_seed(0)
	output_mask = torch.zeros(4, dtype=bool)
	output_mask[3] = True

	model = SumModel()
	y = [[10, 10000, -26, 100]]

	X_hat = greedy_substitution(model, X, y, device='cpu', output_mask=output_mask,
		max_iter=101)

	assert all(X_hat.sum(dim=(0, 2)) == torch.tensor([0, 0, 0, 100]))


def test_greedy_substitution_reverse_complement(X):
	torch.manual_seed(0)
	output_mask = torch.zeros(4, dtype=bool)
	output_mask[3] = True

	model = SmallDeepSEA()
	y = [[10]]

	X_hat = greedy_substitution(model, X, y, ['A', 'C'], device='cpu', max_iter=101,
		reverse_complement=False)

	n_count = X.sum(dim=(0, 2))
	n_hat_count = X_hat.sum(dim=(0, 2))
	
	assert n_count[0] < n_hat_count[0]
	assert n_count[1] < n_hat_count[1]
	assert n_count[2] > n_hat_count[2]
	assert n_count[3] > n_hat_count[3]


def test_greedy_substitution_no_y(X):
	torch.manual_seed(0)
	model = SmallDeepSEA()
	y = [[10]]

	X_hat = greedy_substitution(model, X, device='cpu', max_iter=15)
	y_hat = predict(model, X_hat, device='cpu')

	assert_array_almost_equal(y_hat, [[-0.01224]])


###


def test_greedy_marginalize(X_marg, motifs):
	torch.manual_seed(0)
	model = SmallDeepSEA()
	y = [[10.0]]

	X_hat = greedy_marginalize(model, X_marg, y, motifs, device='cpu', max_iter=1)

	assert X_hat.shape == (4, 12)
	assert X_hat.dtype == torch.int8
	assert X_hat.sum(dim=0).min() == 1
	assert all(torch.unique(X_hat) == torch.tensor([0, 1], dtype=torch.int8))





# kmers.py
# Contact: Jacob Schreiber <jmschreiber91@gmail.com>

import numpy
import torch
import scipy
import numpy as np

from numba import njit, prange
import numba

key_type = numba.types.int64
value_type = numba.types.float64


def kmers(X, k, scores=None):
	"""Extract all k-mers found in a sequence, optionally weighted by a score.

	This function will count the number of k-mers found in each sequence and
	return a feature matrix. If `score` is provided, the counts will be weighted
	by the sum of the score across the k positions where the k-mer resides. If
	`score` is not provided, the count will just be 1 for each instance of the
	k-mer in the sequence.


	Parameters
	----------
	X: torch.Tensor, shape=(-1, len(alphabet), sequence_length)
		A one-hot encoded set of sequences.

	k: int
		The size of the k-mers to consider.
	

	Returns
	-------
	X_kmers: torch.Tensor, shape=(-1, n_kmers)
		A featurization where each row is an example from the original set of
		sequences and each column is a k-mer that could be in the sequence.
	"""

	n = X.shape[1]

	X = X.type(torch.int32)
	w = torch.arange(n).repeat(k, 1).T * n ** torch.arange(k)
	w = w[None, :, :].type(torch.int32).to(X.device)
	idxs = torch.nn.functional.conv1d(X, w).type(torch.int64)[:, 0]

	if scores is not None:
		scores = scores.unsqueeze(1).type(torch.float32)
		ws = torch.ones(1, 1, k, dtype=torch.float32)
		score_ = torch.nn.functional.conv1d(scores, ws)[:, 0]
	else:
		score_ = torch.ones(1, dtype=torch.float32).expand_as(idxs)

	X_kmers = torch.zeros((X.shape[0], n**k))
	X_kmers.scatter_add_(1, idxs, score_)
	return X_kmers


@njit(parallel=True)
def _fast_extract_gkmers(X, min_k, max_k, max_gap, max_len, max_entries):
	nx = X.shape[0]
	keys = np.zeros((nx, max_entries), dtype='int64')
	scores = np.zeros((nx, max_entries), dtype='float64')

	for xi in prange(nx):
		n = X.shape[1]
		gkmer_attrs = numba.typed.Dict.empty(key_type=key_type, 
			value_type=value_type)

		last_k_gkmers = []
		last_k_gkmers_attrs = []
		last_k_gkmers_hashes = []

		for i in range(n):
			base = int(X[xi, i, 1])
			attr = X[xi, i, 2]

			last_k_gkmers.append(np.array([i], dtype='int32'))
			last_k_gkmers_attrs.append(np.array([attr], dtype='float64'))
			last_k_gkmers_hashes.append(np.array([base+1], dtype='int64'))

		for k in range(2, max_k+1):
			for j in range(n):
				start_position = X[xi, j, 0]

				gkmers_ = []
				gkmer_attrs_ = []
				gkmer_hashes_ = []

				for i in range(j+1, n):
					position = X[xi, i, 0]
					base = int(X[xi, i, 1])
					attr = X[xi, i, 2]

					if (position - start_position) >= max_len:
						break

					for g in range(len(last_k_gkmers[j])):
						gkmer = last_k_gkmers[j][g]
						gkmer_attr = last_k_gkmers_attrs[j][g]
						gkmer_hash = last_k_gkmers_hashes[j][g]

						last_position = X[xi, gkmer, 0]
						if last_position >= position:
							break

						if (position - last_position) > max_gap:
							continue

						diff = int(position - last_position - 1)
						length = int(position - start_position)

						new_gkmer_hash = gkmer_hash + (base+1) * (5 ** length)
						new_gkmer_attr = gkmer_attr + attr
						
						gkmers_.append(i)
						gkmer_attrs_.append(new_gkmer_attr)
						gkmer_hashes_.append(new_gkmer_hash)

						if k >= min_k:
							gkmer_attrs[new_gkmer_hash] = gkmer_attrs.get(
								new_gkmer_hash, 0) + new_gkmer_attr / k

				if len(gkmers_) == 0:
					last_k_gkmers[j] = np.zeros(0, dtype='int32')
					last_k_gkmers_attrs[j] = np.zeros(0, dtype='float64')
					last_k_gkmers_hashes[j] = np.zeros(0, dtype='int64')
				else:
					last_k_gkmers[j] = np.array(gkmers_, dtype='int32')
					last_k_gkmers_attrs[j] = np.array(gkmer_attrs_, 
						dtype='float64')
					last_k_gkmers_hashes[j] = np.array(gkmer_hashes_, 
						dtype='int64')
		
		ny = len(gkmer_attrs)
		keys_ = np.empty(ny, dtype='int64')
		scores_ = np.empty(ny, dtype='float64')

		for i, key in enumerate(gkmer_attrs.keys()):
			keys_[i] = key
			scores_[i] = gkmer_attrs[key]

		idxs = np.argsort(-np.abs(scores_), kind='mergesort')[:max_entries]

		keys[xi] = keys_[idxs]
		scores[xi] = scores_[idxs]

	return keys, scores



def gapped_kmers(X, scores=None, min_k=4, max_k=8, max_gap=2, max_len=10, 
	max_gkmers=10, max_pos=None):
	"""Extract gapped k-mers from sequences and optionally scores.

	This function will extract the gapped k-mers from a set of sequences that
	are one-hot encoded and optionally scores for each position. Without scores,
	this function will return a sparse matrix where each row is an example,
	each column is a gapped k-mer, and the value is the count of that gapped
	k-mer in the data. With scores, the shape will be the same but the value 
	will be the sum of the scores across positions in the gapped k-mers.


	Parameters
	----------
	X: torch.Tensor, shape=(-1, len(alphabet), sequence_length)
		A one-hot encoded set of sequences.

	attr: torch.Tensor with shape=(-1, sequence_length) or None, optional
		A corresponding set of attribution values to use. If None, return counts
		instead of sum of attribution values. Default is None.

	min_k: int, optional
		The minimum number of non-gaps in the k-mer. Default is 4.

	max_k: int, optional
		The maximum number of non-gaps in the k-mer. Default is 8.

	max_gap: int, optional
		The maximum number of gaps in the k-mer. Default is 2.

	max_len: int, optional
		The maximum length of the k-mer, as in, the total number of gaps and
		non-gap characters. Default is 10.

	max_gkmers: int, optional
		The maximum number of gapped k-mers to return.

	top_n_gkmers: int, optional
		..


	Returns
	-------
	gkmers: scipy.sparse.csr_matrix
		A sparse matrix containing either counts or score sums for each kmer
		found.
	"""


	X_idxs = X.argmax(axis=1)
	if not scores:
		scores = torch.ones_like(X_idxs)

	if max_pos is None:
		max_pos = X.shape[-1]

	X_scores = numpy.zeros((X.shape[0], max_pos, 3))
	for i in range(X.shape[0]):
		if scores is not None:
			score_idxs = numpy.argsort(-scores[i])[:max_pos].sort().values
		else:
			score_idxs = numpy.arange(X.shape[-1])

		X_scores[i, :, 0] = score_idxs
		X_scores[i, :, 1] = X_idxs[i, score_idxs]
		X_scores[i, :, 2] = scores[i, score_idxs]

	gkmers, gkmer_scores = _fast_extract_gkmers(X_scores, min_k=min_k, max_k=max_k, 
		max_gap=max_gap, max_len=max_len, max_entries=max_gkmers)
	
	row_idxs = numpy.repeat(range(gkmers.shape[0]), gkmers.shape[1])
	csr_mat = scipy.sparse.csr_matrix((gkmer_scores.flatten(), 
		(row_idxs, gkmers.flatten())), shape=(len(gkmers), 5**max_len))

	return csr_mat


# test_pisa.py
# Contact: Jacob Schreiber <jmschreiber91@gmail.com>

import torch
torch.use_deterministic_algorithms(True, warn_only=True)
torch.manual_seed(0)


import pytest
import warnings

from tangermeme.utils import one_hot_encode
from tangermeme.utils import random_one_hot

from tangermeme.ersatz import substitute
from tangermeme.ersatz import shuffle
from tangermeme.ersatz import dinucleotide_shuffle

from tangermeme.deep_lift_shap import deep_lift_shap
from tangermeme.deep_lift_shap import hypothetical_attributions

from tangermeme.pisa import pisa

from .toy_models import SumModel
from .toy_models import FlattenDense
from .toy_models import Conv1
from .toy_models import Scatter
from .toy_models import ConvDense
from .toy_models import ConvPoolDense
from .toy_models import SmallDeepSEA

from numpy.testing import assert_raises
from numpy.testing import assert_array_almost_equal


@pytest.fixture
def X():
	X_ = random_one_hot((2, 4, 100), random_state=0).type(torch.float32)
	X_ = substitute(X_, "ACGTACGT")
	return X_


@pytest.fixture
def references(X):
	return shuffle(X, n=5, random_state=0)


class LambdaWrapper(torch.nn.Module):
	"""Wrapper that runs a given forward function instead of the default.

	Several of the classes in toy_models.py return multiple outputs but the
	attributions from deep_lift_shap require that there's only one output per
	example to explain. This class helps overcome the issues with having
	multiple outputs by slicing out the output we're interested in.


	Parameters
	----------
	model: torch.nn.Module
		A PyTorch model that we want to use.

	forward: function
		A function that takes in a model and a batch of sequences and returns
		some output. Usually this is just running the forward function of the
		model and then slicing out an output.
	"""

	def __init__(self, model, forward):
		super(LambdaWrapper, self).__init__()
		self.model = model
		self._forward = forward

	def forward(self, X, *args):
		return self._forward(self.model, X, *args)


def test_pisa(X):
	torch.manual_seed(0)
	model = Conv1()

	X = X[:, :, :15]
	X_attr = pisa(model, X, device='cpu', n_shuffles=3, 
		random_state=0, batch_size=4)

	assert X_attr.shape == (2, 9, 4, 15)
	assert X_attr.dtype == torch.float32
	
	assert_array_almost_equal(X_attr[:, :2, :, :5], [
		[[[ 0.0000,  0.0000, -0.0000,  0.0103, -0.0000],
          [-0.0000,  0.0000,  0.0339, -0.0000, -0.0000],
          [ 0.0000,  0.0000,  0.0000, -0.0000,  0.0000],
          [ 0.0000,  0.0000, -0.0000, -0.0000,  0.0000]],

         [[ 0.0000, -0.0000,  0.0000,  0.0187,  0.0000],
          [ 0.0000, -0.0000,  0.0202,  0.0000, -0.0000],
          [ 0.0000,  0.0000,  0.0000,  0.0000, -0.0000],
          [ 0.0000,  0.0000, -0.0000, -0.0000,  0.0000]]],


        [[[ 0.0000,  0.0000,  0.0000,  0.0000, -0.0656],
          [ 0.0000, -0.0000,  0.0000, -0.0000, -0.0000],
          [ 0.0000,  0.0000,  0.0000, -0.0000,  0.0000],
          [ 0.0000, -0.0000, -0.0000, -0.0103,  0.0000]],

         [[ 0.0000, -0.0000,  0.0000,  0.0000,  0.0207],
          [ 0.0000, -0.0000, -0.0000,  0.0000, -0.0000],
          [ 0.0000,  0.0000, -0.0000,  0.0000, -0.0000],
          [ 0.0000, -0.0000, -0.0000, -0.0187, -0.0000]]]], 4)


def test_pisa_deep_lift_shap(X):
	torch.manual_seed(0)
	model = Conv1()

	references = dinucleotide_shuffle(X, n=3)

	X_attr0 = pisa(model, X, device='cpu', references=references)
	X_attr1 = deep_lift_shap(model, X, device='cpu', references=references, 
		target=0)
	X_attr2 = deep_lift_shap(model, X, device='cpu', references=references,
		target=28)
	X_attr3 = deep_lift_shap(model, X, device='cpu', references=references,
		target=-1)

	assert_array_almost_equal(X_attr0[:, 0], X_attr1)
	assert_array_almost_equal(X_attr0[:, 28], X_attr2)
	assert_array_almost_equal(X_attr0[:, -1], X_attr3)


def test_pisa_convergence(X):
	torch.manual_seed(0)
	model = SmallDeepSEA()

	with warnings.catch_warnings():
		warnings.simplefilter("error", category=RuntimeWarning)

		pisa(model, X, device='cpu', n_shuffles=3, random_state=0,
			warning_threshold=1e-7)

		assert_raises(RuntimeWarning, deep_lift_shap, model, X, 
			device='cpu', n_shuffles=3, random_state=0, warning_threshold=1e-10)


def test_pisa_hypothetical(X):
	torch.manual_seed(0)
	model = Conv1()

	X_attr = pisa(model, X, hypothetical=True, device='cpu', 
		random_state=0)

	assert X_attr.shape == (2, 94, 4, 100)
	assert X_attr.dtype == torch.float32

	assert_array_almost_equal(X_attr[:, :3, :, :4], [
		[[[ 0.0000,  0.0482,  0.0029,  0.0731],
          [-0.0195,  0.0110,  0.0487, -0.0715],
          [ 0.0362,  0.0216,  0.0686, -0.1113],
          [ 0.0155, -0.0496, -0.0531,  0.0421]],

         [[ 0.0000, -0.0023,  0.0568, -0.0045],
          [ 0.0000, -0.0218,  0.0196,  0.0412],
          [ 0.0000,  0.0340,  0.0302,  0.0612],
          [ 0.0000,  0.0132, -0.0411, -0.0605]],

         [[ 0.0000,  0.0000, -0.0103,  0.0426],
          [ 0.0000,  0.0000, -0.0298,  0.0055],
          [ 0.0000,  0.0000,  0.0259,  0.0160],
          [ 0.0000,  0.0000,  0.0052, -0.0552]]],


        [[[ 0.0195,  0.0344,  0.0001,  0.1124],
          [ 0.0000, -0.0028,  0.0459, -0.0322],
          [ 0.0557,  0.0077,  0.0658, -0.0720],
          [ 0.0350, -0.0635, -0.0559,  0.0814]],

         [[ 0.0000, -0.0221,  0.0519, -0.0224],
          [ 0.0000, -0.0416,  0.0147,  0.0233],
          [ 0.0000,  0.0141,  0.0253,  0.0432],
          [ 0.0000, -0.0066, -0.0459, -0.0784]],

         [[ 0.0000,  0.0000, -0.0095,  0.0420],
          [ 0.0000,  0.0000, -0.0290,  0.0049],
          [ 0.0000,  0.0000,  0.0267,  0.0154],
          [ 0.0000,  0.0000,  0.0059, -0.0558]]]
	], 4)


def test_pisa_independence():
	X_ = random_one_hot((12, 4, 25), random_state=0).type(torch.float32)
	X = substitute(X_, "ACGTACGT")

	torch.manual_seed(0)
	model = Conv1()

	X_attr = pisa(model, X, device='cpu', random_state=0)
	X_attr0 = pisa(model, X[0:1], device='cpu', random_state=0)
	X_attr1 = pisa(model, X[5:6], device='cpu', random_state=0)
	X_attr2 = pisa(model, X[8:10], device='cpu', random_state=0)
	X_attr3 = pisa(model, X[0:10], device='cpu', random_state=0)

	assert_array_almost_equal(X_attr[0:1], X_attr0)
	assert_array_almost_equal(X_attr[5:6], X_attr1)
	assert_array_almost_equal(X_attr[8:10], X_attr2)
	assert_array_almost_equal(X_attr[0:10], X_attr3)

	assert_raises(AssertionError, assert_array_almost_equal, X_attr0, X_attr1)
	assert_raises(AssertionError, assert_array_almost_equal, X_attr2, 
		X_attr3[:2])


def test_pisa_random_state(X):
	torch.manual_seed(0)
	model = Conv1()

	X_attr0 = pisa(model, X, device='cpu', random_state=0)
	X_attr1 = pisa(model, X, device='cpu', random_state=1)
	X_attr2 = pisa(model, X, device='cpu', random_state=2)

	assert_raises(AssertionError, assert_array_almost_equal, X_attr0, X_attr1)
	assert_raises(AssertionError, assert_array_almost_equal, X_attr0, X_attr2)


def test_pisa_reference_tensor(X):
	torch.manual_seed(0)
	model = Conv1()

	references = shuffle(X, n=20, random_state=0)

	X_attr0 = pisa(model, X, references=references, device='cpu', random_state=0)
	X_attr1 = pisa(model, X, references=references, device='cpu', random_state=1)
	X_attr2 = pisa(model, X, references=references, device='cpu', random_state=2)

	assert_array_almost_equal(X_attr0, X_attr1)
	assert_array_almost_equal(X_attr0, X_attr2)


def test_pisa_batch_size(X):
	torch.manual_seed(0)
	model = Conv1()
	X = X[:1, :, :15]

	X_attr0 = pisa(model, X, device='cpu', random_state=0, n_shuffles=3)
	X_attr1 = pisa(model, X, batch_size=1, device='cpu', random_state=0, n_shuffles=3)
	X_attr2 = pisa(model, X, batch_size=1000, device='cpu', random_state=0, n_shuffles=3)

	assert_array_almost_equal(X_attr0, X_attr1)
	assert_array_almost_equal(X_attr0, X_attr2)


def test_pisa_n_shuffles(X):
	torch.manual_seed(0)
	model = Conv1()

	X = X[:, :, :30]

	X_attr0 = deep_lift_shap(model, X, n_shuffles=1, device='cpu', 
		random_state=0)
	X_attr1 = deep_lift_shap(model, X, n_shuffles=1, batch_size=1, device='cpu', 
		random_state=0)
	X_attr2 = deep_lift_shap(model, X, n_shuffles=30, batch_size=100000, 
		device='cpu', random_state=2)
	X_attr3 = deep_lift_shap(model, X, n_shuffles=30, batch_size=1, 
		device='cpu', random_state=2)

	assert_array_almost_equal(X_attr0, X_attr1)
	assert_array_almost_equal(X_attr2, X_attr3)
	assert_raises(AssertionError, assert_array_almost_equal, X_attr0, X_attr3)


def test_deep_lift_shap_shuffle_ordering(X):
	torch.manual_seed(0)
	model = Conv1()
	X = X[:1]

	references = dinucleotide_shuffle(X, n=1, random_state=0)

	X_attr0 = pisa(model, X, n_shuffles=1, device='cpu', random_state=0)
	X_attr1 = pisa(model, X, device='cpu', references=references)

	assert_array_almost_equal(X_attr0, X_attr1)


def test_pisa_raw_output(X):
	torch.manual_seed(0)
	model = Conv1()

	X_attr0, refs = pisa(model, X, device='cpu', raw_outputs=True, 
		random_state=0, return_references=True)
	X_attr1 = pisa(model, X, device='cpu', random_state=0)

	assert refs.shape == (2, 20, 4, 100)
	assert X_attr0.shape == (2, 20, 94, 4, 100)
	assert X_attr1.shape == (2, 94, 4, 100)

	assert_array_almost_equal(X_attr0[:, :2, :3, :, :3], [
		[[[[ 0.0054,  0.0253,  0.0155],
           [-0.0141, -0.0119,  0.0612],
           [ 0.0417, -0.0013,  0.0811],
           [ 0.0209, -0.0726, -0.0406]],

          [[ 0.0000,  0.0054,  0.0253],
           [ 0.0000, -0.0141, -0.0119],
           [ 0.0000,  0.0417, -0.0013],
           [ 0.0000,  0.0209, -0.0726]],

          [[ 0.0000,  0.0000,  0.0054],
           [ 0.0000,  0.0000, -0.0141],
           [ 0.0000,  0.0000,  0.0417],
           [ 0.0000,  0.0000,  0.0209]]],


         [[[ 0.0054,  0.0253,  0.0155],
           [-0.0141, -0.0119,  0.0612],
           [ 0.0417, -0.0013,  0.0811],
           [ 0.0209, -0.0726, -0.0406]],

          [[ 0.0000,  0.0054,  0.0253],
           [ 0.0000, -0.0141, -0.0119],
           [ 0.0000,  0.0417, -0.0013],
           [ 0.0000,  0.0209, -0.0726]],

          [[ 0.0000,  0.0000,  0.0054],
           [ 0.0000,  0.0000, -0.0141],
           [ 0.0000,  0.0000,  0.0417],
           [ 0.0000,  0.0000,  0.0209]]]],



        [[[[ 0.0054,  0.0253,  0.0155],
           [-0.0141, -0.0119,  0.0612],
           [ 0.0417, -0.0013,  0.0811],
           [ 0.0209, -0.0726, -0.0406]],

          [[ 0.0000,  0.0054,  0.0253],
           [ 0.0000, -0.0141, -0.0119],
           [ 0.0000,  0.0417, -0.0013],
           [ 0.0000,  0.0209, -0.0726]],

          [[ 0.0000,  0.0000,  0.0054],
           [ 0.0000,  0.0000, -0.0141],
           [ 0.0000,  0.0000,  0.0417],
           [ 0.0000,  0.0000,  0.0209]]],


         [[[ 0.0054,  0.0253,  0.0155],
           [-0.0141, -0.0119,  0.0612],
           [ 0.0417, -0.0013,  0.0811],
           [ 0.0209, -0.0726, -0.0406]],

          [[ 0.0000,  0.0054,  0.0253],
           [ 0.0000, -0.0141, -0.0119],
           [ 0.0000,  0.0417, -0.0013],
           [ 0.0000,  0.0209, -0.0726]],

          [[ 0.0000,  0.0000,  0.0054],
           [ 0.0000,  0.0000, -0.0141],
           [ 0.0000,  0.0000,  0.0417],
           [ 0.0000,  0.0000,  0.0209]]]]
	], 4)

	X_attr2 = hypothetical_attributions(
		(X_attr0.reshape(-1, 4, 100),), 
		(X.repeat_interleave(20*94, dim=0),), 
		(refs.reshape(-1, 4, 100).repeat_interleave(94, dim=0),)
	)[0]

	X_attr2 = X_attr2.reshape(X.shape[0], 20, 94, 4, 100)
	X_attr2 = torch.mean(X_attr2 * X.unsqueeze(1).unsqueeze(1), dim=1)

	assert_array_almost_equal(X_attr2, X_attr1, 4)


def test_pisa_return_references(X):
	torch.manual_seed(0)
	model = Conv1()

	attr, refs = pisa(model, X, n_shuffles=1, return_references=True,
		device='cpu', random_state=0)

	assert attr.shape == (2, 94, 4, 100)
	assert refs.shape == (2, 1, 4, 100)
	assert refs.dtype == torch.float32
	assert refs[:, 0].sum(dim=1).max() == 1

	assert_array_almost_equal(refs[:4, :, :, :10], [
		[[[1., 0., 0., 0., 0., 0., 0., 0., 1., 0.],
          [0., 1., 0., 0., 1., 0., 0., 0., 0., 0.],
          [0., 0., 1., 0., 0., 0., 0., 0., 0., 1.],
          [0., 0., 0., 1., 0., 1., 1., 1., 0., 0.]]],

        [[[0., 0., 0., 1., 1., 0., 0., 0., 1., 0.],
          [1., 0., 0., 0., 0., 0., 0., 0., 0., 0.],
          [0., 0., 0., 0., 0., 0., 1., 0., 0., 0.],
          [0., 1., 1., 0., 0., 1., 0., 1., 0., 1.]]]])


	_, refs2 = pisa(model, X, n_shuffles=3, return_references=True,
		device='cpu', random_state=0)

	assert_array_almost_equal(refs, refs2[:, 0:1])


def test_pisa_args(X):
	torch.manual_seed(0)
	model = FlattenDense(n_outputs=1)
	alpha = torch.randn(16, 1)
	beta = torch.randn(16, 1)

	X_attr0 = pisa(model, X, device='cpu', random_state=0)[:, 0]
	X_attr1 = pisa(model, X, args=(alpha,), device='cpu', 
		random_state=0)[:, 0]
	X_attr2 = pisa(model, X, args=(alpha, beta), device='cpu', 
		random_state=0)[:, 0]

	assert X.shape == X_attr0.shape
	assert X.shape == X_attr1.shape
	assert X.shape == X_attr2.shape

	assert_array_almost_equal(X_attr0, X_attr1)
	assert_raises(AssertionError, assert_array_almost_equal, X_attr0, X_attr2)

	assert_array_almost_equal(X_attr2[:2, :, :10], [
		[[ 0.0000,  0.0000, -0.0000, -0.0209, -0.0000,  0.0000,  0.0000,
           0.0000, -0.0000,  0.0000],
         [ 0.0000,  0.0000,  0.0080,  0.0000, -0.0000,  0.0000, -0.0000,
          -0.0000, -0.0081,  0.0000],
         [ 0.0000,  0.0000, -0.0000, -0.0000, -0.0000,  0.0000, -0.0000,
           0.0000,  0.0000,  0.0000],
         [-0.0000, -0.0160,  0.0000,  0.0000,  0.0245, -0.0336,  0.0200,
           0.0050,  0.0000, -0.0085]],

        [[-0.0000, -0.0000, -0.0090, -0.0000, -0.0092,  0.0000,  0.0087,
           0.0000, -0.0000,  0.0079],
         [ 0.0000,  0.0000,  0.0000,  0.0000, -0.0000, -0.0000, -0.0000,
          -0.0000, -0.0000,  0.0000],
         [-0.0000,  0.0018, -0.0000, -0.0000, -0.0000,  0.0322, -0.0000,
           0.0000,  0.0000,  0.0000],
         [-0.0000, -0.0000,  0.0000,  0.0265,  0.0000, -0.0000,  0.0000,
          -0.0006,  0.0063, -0.0000]]
	], 4)


def test_pisa_raises(references):
	X_ = random_one_hot((16, 4, 100), random_state=0).type(torch.float32)
	X = substitute(X_, "ACGTACGT")
	
	torch.manual_seed(0)
	model = FlattenDense(n_outputs=1)
	alpha = torch.randn(16, 1)
	beta = torch.randn(16, 1)

	assert_raises(ValueError, pisa, model, X[0], device='cpu')
	assert_raises(ValueError, pisa, model, X.unsqueeze(1), 
		device='cpu')
	assert_raises(RuntimeError, pisa, model, X, n_shuffles=0, 
		device='cpu')
	assert_raises(ValueError, pisa, model, X[0], device='cpu')

	assert_raises(IndexError, pisa, model, X, args=(alpha[:10],),
		device='cpu')
	assert_raises(IndexError, pisa, model, X, args=(alpha, beta[:3]),
		device='cpu')
	assert_raises(IndexError, pisa, model, X, args=(alpha[:5], 
		beta[:3]), device='cpu')
	assert_raises(IndexError, pisa, model, X, args=(alpha, beta[:3]),
		device='cpu')
	
	assert_raises(ValueError, pisa, model, X, 
		references=references[:10], device='cpu')
	assert_raises(ValueError, pisa, model, X, 
		references=references[:, :, :2], device='cpu')
	assert_raises(ValueError, pisa, model, X, 
		references=references[:, :, :, :10], device='cpu')


### Test a bunch of different models with different configurations/operations


def test_pisa_flattendense(X, references):
	torch.manual_seed(0)
	model = FlattenDense(n_outputs=1)

	with warnings.catch_warnings():
		warnings.simplefilter("error", category=RuntimeWarning)

		X_attr = pisa(model, X, device='cpu', references=references, 
			warning_threshold=1e-5)

		assert_raises(RuntimeWarning, pisa, model, X, device='cpu', 
			random_state=0, warning_threshold=1e-10)

	X_attr2 = deep_lift_shap(model, X, device='cpu', references=references)
	
	assert X_attr.dtype == torch.float32
	assert_array_almost_equal(X_attr[:, 0], X_attr2)


def test_pisa_flattendense_n_outputs(X):
	model = FlattenDense(n_outputs=1)
	X_attr = pisa(model, X, device='cpu')
	assert X_attr.shape == (2, 1, 4, 100)

	model = FlattenDense(n_outputs=4)
	X_attr = pisa(model, X, device='cpu')
	assert X_attr.shape == (2, 4, 4, 100)

	model = FlattenDense(n_outputs=12)
	X_attr = pisa(model, X, device='cpu')
	assert X_attr.shape == (2, 12, 4, 100)


def test_pisa_convdense_dense_wrapper(X, references):
	torch.manual_seed(0)
	model = LambdaWrapper(ConvDense(n_outputs=1), lambda model, X: model(X)[1])

	with warnings.catch_warnings():
		warnings.simplefilter("error", category=RuntimeWarning)

		X_attr = pisa(model, X, device='cpu', references=references, 
			warning_threshold=1e-5)

		assert_raises(RuntimeWarning, deep_lift_shap, model, X, device='cpu', 
			random_state=0, warning_threshold=1e-8)

	X_attr2 = deep_lift_shap(model, X, device='cpu', references=references)
	assert_array_almost_equal(X_attr[:, 0], X_attr2)


def test_pisa_convdense_conv_wrapper(X, references):
	torch.manual_seed(0)
	model = LambdaWrapper(ConvDense(n_outputs=1), 
		lambda model, X: model(X)[0].sum(dim=(-1, -2)).unsqueeze(-1))

	with warnings.catch_warnings():
		warnings.simplefilter("error", category=RuntimeWarning)

		X_attr = pisa(model, X, device='cpu', references=references, 
			warning_threshold=1e-4)

		assert_raises(RuntimeWarning, deep_lift_shap, model, X, device='cpu', 
			random_state=0, warning_threshold=1e-8)

	X_attr2 = deep_lift_shap(model, X, device='cpu', references=references)
	assert_array_almost_equal(X_attr[:, 0], X_attr2)


### Now test some custom models with weird architectures just to check


class TorchSum(torch.nn.Module):
	def __init__(self):
		super(TorchSum, self).__init__()

	def forward(self, X):
		if len(X.shape) == 2:
			return torch.sum(X, dim=-1, keepdims=True)
		else:
			return torch.sum(X, dim=(-1, -2)).unsqueeze(-1)


def test_pisa_linear(X):
	torch.manual_seed(0)

	model = torch.nn.Sequential(
		torch.nn.Flatten(),
		torch.nn.Linear(400, 5),
		torch.nn.ReLU(),
		TorchSum()
	)

	with warnings.catch_warnings():
		warnings.simplefilter("error", category=RuntimeWarning)
		X_attr = pisa(model, X, device='cpu', random_state=0, 
			warning_threshold=1e-5)


def test_pisa_linear_bias(X):
	torch.manual_seed(0)

	model = torch.nn.Sequential(
		torch.nn.Flatten(),
		torch.nn.Linear(400, 5, bias=False),
		torch.nn.ReLU(),
		TorchSum()
	)

	with warnings.catch_warnings():
		warnings.simplefilter("error", category=RuntimeWarning)
		X_attr = pisa(model, X, device='cpu', random_state=0, 
			warning_threshold=1e-5)


def test_pisa_conv(X):
	torch.manual_seed(0)

	model = torch.nn.Sequential(
		torch.nn.Conv1d(4, 8, (5,)),
		torch.nn.ReLU(),
		TorchSum()
	)

	with warnings.catch_warnings():
		warnings.simplefilter("error", category=RuntimeWarning)
		X_attr = pisa(model, X, device='cpu', random_state=0, 
			warning_threshold=1e-5)


def test_pisa_conv_dilated(X):
	torch.manual_seed(0)

	model = torch.nn.Sequential(
		torch.nn.Conv1d(4, 8, (5,), dilation=5),
		torch.nn.ReLU(),
		TorchSum()
	)

	with warnings.catch_warnings():
		warnings.simplefilter("error", category=RuntimeWarning)
		X_attr = pisa(model, X, device='cpu', random_state=0, 
			warning_threshold=1e-5)


def test_pisa_conv_stride(X):
	torch.manual_seed(0)

	model = torch.nn.Sequential(
		torch.nn.Conv1d(4, 8, (5,), stride=5),
		torch.nn.ReLU(),
		TorchSum()
	)

	with warnings.catch_warnings():
		warnings.simplefilter("error", category=RuntimeWarning)
		X_attr = pisa(model, X, device='cpu', random_state=0, 
			warning_threshold=1e-5)


def test_pisa_conv_bias(X):
	torch.manual_seed(0)

	model = torch.nn.Sequential(
		torch.nn.Conv1d(4, 8, (5,), bias=False),
		torch.nn.ReLU(),
		TorchSum()
	)

	with warnings.catch_warnings():
		warnings.simplefilter("error", category=RuntimeWarning)
		X_attr = pisa(model, X, device='cpu', random_state=0, 
			warning_threshold=1e-4)


def test_pisa_conv_padding(X):
	torch.manual_seed(0)

	model = torch.nn.Sequential(
		torch.nn.Conv1d(4, 8, (5,), padding=5),
		torch.nn.ReLU(),
		TorchSum()
	)

	with warnings.catch_warnings():
		warnings.simplefilter("error", category=RuntimeWarning)
		X_attr = pisa(model, X, device='cpu', random_state=0, 
			warning_threshold=1e-4)


def test_pisa_conv_padding_same(X):
	torch.manual_seed(0)

	model = torch.nn.Sequential(
		torch.nn.Conv1d(4, 8, (5,), padding='same'),
		torch.nn.Flatten(),
		torch.nn.ReLU(),
		torch.nn.Linear(800, 1)
	)

	with warnings.catch_warnings():
		warnings.simplefilter("error", category=RuntimeWarning)
		X_attr = pisa(model, X, device='cpu', random_state=0, 
			warning_threshold=1e-5)


def test_pisa_max_pool(X):
	torch.manual_seed(0)

	model = torch.nn.Sequential(
		torch.nn.MaxPool1d(4),
		TorchSum()
	)

	with warnings.catch_warnings():
		warnings.simplefilter("error", category=RuntimeWarning)
		X_attr = pisa(model, X, device='cpu', random_state=0, 
			warning_threshold=1e-5)


def test_pisa_conv_relu_pool(X):
	torch.manual_seed(0)

	model = torch.nn.Sequential(
		torch.nn.Conv1d(4, 8, (5,)),
		torch.nn.ReLU(),
		torch.nn.MaxPool1d(4),
		TorchSum()
	)

	with warnings.catch_warnings():
		warnings.simplefilter("error", category=RuntimeWarning)
		X_attr = pisa(model, X, device='cpu', random_state=0, 
			warning_threshold=1e-5)


def test_pisa_conv_tanh_pool(X):
	torch.manual_seed(0)

	model = torch.nn.Sequential(
		torch.nn.Conv1d(4, 8, (5,)),
		torch.nn.Tanh(),
		torch.nn.MaxPool1d(4),
		TorchSum()
	)

	with warnings.catch_warnings():
		warnings.simplefilter("error", category=RuntimeWarning)
		X_attr = pisa(model, X, device='cpu', random_state=0, 
			warning_threshold=1e-5)


def test_pisa_conv_elu_pool(X):
	torch.manual_seed(0)

	model = torch.nn.Sequential(
		torch.nn.Conv1d(4, 8, (5,)),
		torch.nn.ELU(),
		torch.nn.MaxPool1d(4),
		TorchSum()
	)

	with warnings.catch_warnings():
		warnings.simplefilter("error", category=RuntimeWarning)
		X_attr = pisa(model, X, device='cpu', random_state=0, 
			warning_threshold=1e-5)


def test_pisa_conv_relu_pool_relu(X):
	torch.manual_seed(0)

	model = torch.nn.Sequential(
		torch.nn.Conv1d(4, 8, (5,)),
		torch.nn.ReLU(),
		torch.nn.MaxPool1d(4),
		torch.nn.ReLU(),
		TorchSum()
	)

	with warnings.catch_warnings():
		warnings.simplefilter("error", category=RuntimeWarning)
		X_attr = pisa(model, X, device='cpu', random_state=0, 
			warning_threshold=1e-5)


def test_pisa_relu_conv_relu_pool_relu(X):
	torch.manual_seed(0)

	model = torch.nn.Sequential(
		torch.nn.ReLU(),
		torch.nn.Conv1d(4, 8, (5,)),
		torch.nn.ReLU(),
		torch.nn.MaxPool1d(4),
		torch.nn.ReLU(),
		TorchSum()
	)

	with warnings.catch_warnings():
		warnings.simplefilter("error", category=RuntimeWarning)
		X_attr = pisa(model, X, device='cpu', random_state=0, 
			warning_threshold=1e-5)


def test_pisa_relu_conv_pool_relu(X):
	torch.manual_seed(0)

	model = torch.nn.Sequential(
		torch.nn.ReLU(),
		torch.nn.Conv1d(4, 8, (5,)),
		torch.nn.MaxPool1d(4),
		torch.nn.ReLU(),
		TorchSum()
	)

	with warnings.catch_warnings():
		warnings.simplefilter("error", category=RuntimeWarning)
		X_attr = pisa(model, X, device='cpu', random_state=0, 
			warning_threshold=1e-5)


def test_pisa_relu_conv_pool_relu_relu(X):
	torch.manual_seed(0)

	model = torch.nn.Sequential(
		torch.nn.ReLU(),
		torch.nn.Conv1d(4, 8, (5,)),
		torch.nn.MaxPool1d(4),
		torch.nn.ReLU(),
		torch.nn.ReLU(),
		TorchSum()
	)

	with warnings.catch_warnings():
		warnings.simplefilter("error", category=RuntimeWarning)
		X_attr = pisa(model, X, device='cpu', random_state=0, 
			warning_threshold=1e-5)


def test_pisa_conv_relu_tanh_pool(X):
	torch.manual_seed(0)

	model = torch.nn.Sequential(
		torch.nn.Conv1d(4, 8, (5,)),
		torch.nn.ReLU(),
		torch.nn.Tanh(),
		torch.nn.MaxPool1d(4),
		TorchSum()
	)

	with warnings.catch_warnings():
		warnings.simplefilter("error", category=RuntimeWarning)
		X_attr = pisa(model, X, device='cpu', random_state=0, 
			warning_threshold=1e-5)


def test_pisa_conv_relu_pool_linear(X):
	torch.manual_seed(0)

	model = torch.nn.Sequential(
		torch.nn.Conv1d(4, 8, (5,)),
		torch.nn.ReLU(),
		torch.nn.MaxPool1d(4),
		torch.nn.Flatten(),
		torch.nn.Linear(192, 1)
	)

	with warnings.catch_warnings():
		warnings.simplefilter("error", category=RuntimeWarning)
		X_attr = pisa(model, X, device='cpu', random_state=0, 
			warning_threshold=1e-5)


def test_pisa_conv_relu_pool_linear_linear(X):
	torch.manual_seed(0)

	model = torch.nn.Sequential(
		torch.nn.Conv1d(4, 8, (5,)),
		torch.nn.ReLU(),
		torch.nn.MaxPool1d(4),
		torch.nn.Flatten(),
		torch.nn.Linear(192, 10),
		torch.nn.Linear(10, 1)
	)

	with warnings.catch_warnings():
		warnings.simplefilter("error", category=RuntimeWarning)
		X_attr = pisa(model, X, device='cpu', random_state=0, 
			warning_threshold=1e-5)


def test_pisa_conv_relu_pool_linear_relu_linear(X):
	torch.manual_seed(0)

	model = torch.nn.Sequential(
		torch.nn.Conv1d(4, 8, (5,)),
		torch.nn.ReLU(),
		torch.nn.MaxPool1d(4),
		torch.nn.Flatten(),
		torch.nn.Linear(192, 10),
		torch.nn.ReLU(),
		torch.nn.Linear(10, 1)
	)

	with warnings.catch_warnings():
		warnings.simplefilter("error", category=RuntimeWarning)
		X_attr = pisa(model, X, device='cpu', random_state=0, 
			warning_threshold=1e-5)



# pisa.py
# Based on the algorithm proposed by Charles McAnany et al.
# https://www.biorxiv.org/content/10.1101/2025.04.07.647613v2
# Contact: Jacob Schreiber <jmschreiber91@gmail.com>


import torch
import warnings

from tqdm import trange
from .ersatz import dinucleotide_shuffle

from tangermeme.predict import predict
from tangermeme.utils import _validate_input

from tangermeme.deep_lift_shap import _nonlinear, _maxpool, _softmax
from tangermeme.deep_lift_shap import _clear_hooks, _register_hooks
from tangermeme.deep_lift_shap import hypothetical_attributions


def pisa(model, X, args=None, batch_size=32, references=dinucleotide_shuffle, 
    n_shuffles=20, return_references=False, hypothetical=False, 
    warning_threshold=0.001, additional_nonlinear_ops=None, 
    print_convergence_deltas=False, raw_outputs=False, device='cuda', 
    random_state=None, verbose=False):
    """An implementation of Pairwise Influence by Sequence Attribution (PISA).
    
    PISA is a method for deciphering how each input in an example influences
    each output. For sequence-based machine learning models this usually means
    how each nucleotide influences the predictions made at each output. This
    can be useful for understanding precisely what parts of a predicted profile
    are influenced by the presence of certain motifs or other changes in
    nucleotide composition.
    
    In practice, PISA involves running DeepLIFT/SHAP once for each output
    position. Given a fixed input and model, one might suspect that the
    Jacobian could be calculated simply to speed up DeepLIFT/SHAP, but this
    seems to not be the case. This is, in part, due to DeepLIFT/SHAP overriding
    some of the gradient operations in ways that do not play nicely with some
    of PyTorch's built-in functions.
    
    
    Parameters
    ----------
    model: torch.nn.Module
        A PyTorch model to use for making predictions. These models can take in
        any number of inputs and make any number of outputs. The additional
        inputs must be specified in the `args` parameter.

    X: torch.tensor, shape=(-1, len(alphabet), length)
        A set of one-hot encoded sequences to calculate attribution values
        for. 

    args: tuple or None, optional
        An optional set of additional arguments to pass into the model. If
        provided, each element in the tuple or list is one input to the model
        and the element must be formatted to be the same batch size as `X`. If
        None, no additional arguments are passed into the forward function.
        Default is None.

    batch_size: int, optional
        The number of sequence-reference pairs to pass through DeepLiftShap at
        a time. Importantly, this is not the number of elements in `X` that
        are processed simultaneously (alongside ALL their references) but the
        total number of `X`-`reference` pairs that are processed. This means
        that if you are in a memory-limited setting where you cannot process
        all references for even a single sequence simultaneously that the
        work is broken down into doing only a few references at a time. Default
        is 32.

    references: func or torch.Tensor, optional
        If a function is passed in, this function is applied to each sequence
        with the provided random state and number of shuffles. This function
        should serve to transform a sequence into some form of signal-null
        background, such as by shuffling it. If a torch.Tensor is passed in,
        that tensor must have shape `(len(X), n_shuffles, *X.shape[1:])`, in
        that for each sequence a number of shuffles are provided. Default is
        the function `dinucleotide_shuffle`. 

    n_shuffles: int, optional
        The number of shuffles to use if a function is given for `references`.
        If a torch.Tensor is provided, this number is ignored. Default is 20.

    return_references: bool, optional
        Whether to return the references that were generated during this
        process. Only use if `references` is not a torch.Tensor. Default is 
        False. 

    hypothetical: bool, optional
        Whether to return attributions for all possible characters at each
        position or only for the character that is actually at the sequence.
        Practically, whether to return the returned attributions from captum
        with the one-hot encoded sequence. Default is False.

    warning_threshold: float, optional
        A threshold on the convergence delta that will always raise a warning
        if the delta is larger than it. Normal deltas are in the range of
        1e-6 to 1e-8. Note that convergence deltas are calculated on the
        gradients prior to the aggr_func being applied to them. Default 
        is 0.001. 

    additional_nonlinear_ops: dict or None, optional
        If additional nonlinear ops need to be added to the dictionary of
        operations that can be handled by DeepLIFT/SHAP, pass a dictionary here
        where the keys are class types and the values are the name of the
        function that handle that sort of class. Make sure that the signature
        matches those of `_nonlinear` and `_maxpool` above. This can also be
        used to overwrite the hard-coded operations by passing in a dictionary
        with overlapping key names. If None, do not add any additional 
        operations. Default is None.

    print_convergence_deltas: bool, optional
        Whether to print the convergence deltas for each example when using
        DeepLiftShap. Default is False.

    raw_outputs: bool, optional
        Whether to return the raw outputs from the method -- in this case,
        the multipliers for each example-reference pair -- or the processed
        attribution values. Default is False.

    device: str or torch.device, optional
        The device to move the model and batches to when making predictions. If
        set to 'cuda' without a GPU, this function will crash and must be set
        to 'cpu'. Default is 'cuda'. 

    random_state: int or None or numpy.random.RandomState, optional
        The random seed to use to ensure determinism. If None, the
        process is not deterministic. Default is None. 

    verbose: bool, optional
        Whether to display a progress bar. Default is False.


    Returns
    -------
    attributions: torch.tensor
        If `raw_outputs=False` (default), the attribution values with shape
        equal to `X`. If `raw_outputs=True`, the multipliers for each example-
        reference pair with shape equal to `(X.shape[0], n_shuffles, X.shape[1],
        X.shape[2])`. 

    references: torch.tensor, optional
        The references used for each input sequence, with the shape
        (n_input_sequences, n_shuffles, 4, length). Only returned if
        `return_references = True`. 
    """    

    _validate_input(X, "X", shape=(-1, -1, -1), ohe=True)

    _NON_LINEAR_OPS = {
        torch.nn.ReLU: _nonlinear,
        torch.nn.ReLU6: _nonlinear,
        torch.nn.RReLU: _nonlinear,
        torch.nn.SELU: _nonlinear,
        torch.nn.CELU: _nonlinear,
        torch.nn.GELU: _nonlinear,
        torch.nn.SiLU: _nonlinear,
        torch.nn.Mish: _nonlinear,
        torch.nn.GLU: _nonlinear,
        torch.nn.ELU: _nonlinear,
        torch.nn.LeakyReLU: _nonlinear,
        torch.nn.Sigmoid: _nonlinear,
        torch.nn.Tanh: _nonlinear,
        torch.nn.Softplus: _nonlinear,
        torch.nn.Softshrink: _nonlinear,
        torch.nn.LogSigmoid: _nonlinear,
        torch.nn.PReLU: _nonlinear,
        torch.nn.MaxPool1d: _maxpool,
        torch.nn.MaxPool2d: _maxpool,
        torch.nn.Softmax: _softmax
    }

    # Misc. set up for overriding operations

    if additional_nonlinear_ops is not None:
        for key, value in additional_nonlinear_ops.items():
            _NON_LINEAR_OPS[key] = value

    model = model.to(device).eval()
    for module in model.modules():
        module._NON_LINEAR_OPS = _NON_LINEAR_OPS

    try:
        model.apply(_register_hooks)
    except Exception as e:
        model.apply(_clear_hooks)
        raise(e)

    # Begin PISA procedure    
    attributions, references_ = [], [] 
    if isinstance(references, torch.Tensor):
        _validate_input(references, "references", shape=(X.shape[0], -1, X.shape[1], 
            X.shape[2]), ohe=True, allow_N=False, ohe_dim=-2)
        n_shuffles = references.shape[1]

    n_outputs = predict(model, X[:1], device=device).shape[-1]

    # Loop over each of the examples
    for i in trange(len(X), disable=not verbose):
        _X = X[i:i+1].to(device).requires_grad_()

        # Either use reference sequences if provided or generate a set of them
        # for this example being analyzed.
        if isinstance(references, torch.Tensor):
            _references = references[i]
        else:
            _references = references(_X.cpu(), n=n_shuffles, 
                random_state=random_state)[0]

        _references = _references.to(device).requires_grad_()
        if return_references:
            references_.append(_references)

        # Pull out the additional arguments for this example, if additional
        # arguments are being provided
        _args = None if args is None else tuple([a[i].to(device) 
            for a in args])

        multipliers = []

        # Loop over all of the shuffles for each example
        for j in range(n_shuffles):
            # For each output explained, the input and the reference will be
            # the same, so extract those single elements and expand them to
            # be the batch size.

            x = _X.expand(batch_size, -1, -1)
            r = _references[j:j+1].expand(batch_size, -1, -1)
            xr = (x[:1] - r[:1])
            
            X_ = torch.cat([x, r])
            
            with torch.autograd.set_grad_enabled(True):
                if _args is not None:
                    _args = (torch.cat([arg, arg]) for arg in _args)
                    y = model(X_, *_args)
                else:
                    y = model(X_)

                _multipliers = []
                edge_size = n_outputs % batch_size

                # Loop over all of the output positions
                for k in range(0, n_outputs, batch_size):
                    b = min(batch_size, n_outputs - k)
 
                    rows = torch.cat([torch.arange(b), torch.arange(b)+batch_size])
                    cols = torch.arange(k, k+b).repeat(2)
                    y_hat = y[rows, cols].sum()
                    
                    multipliers_ = torch.autograd.grad(y_hat, x, retain_graph=True)[0]
                    
                    # Check that the prediction-difference-from-reference is equal to
                    # the sum of the attributions                    
                    output_diff = torch.sub(*torch.chunk(y[rows, cols], 2))
                    input_diff = torch.sum(xr * multipliers_[:b], dim=(1, 2))
                                       
                    convergence_deltas = abs(output_diff - input_diff)
                    if torch.any(convergence_deltas > warning_threshold):
                        warnings.warn("Convergence deltas too high: " +   
                            str(convergence_deltas), RuntimeWarning)

                    if print_convergence_deltas:
                        print(convergence_deltas)
                    
                    if not raw_outputs:
                        multipliers_ = hypothetical_attributions(
                            (multipliers_[:b],), 
                            (x[:b],), 
                            (r[:b],)
                        )[0].cpu().detach()    
                                       
                    _multipliers.append(multipliers_[:b])                  

            # Discard the accumulated graph
            torch.autograd.grad(y[:1, :1].sum(), x, retain_graph=False)[0]

            # Keep the multpliers across each output
            multipliers.append(torch.cat(_multipliers, dim=0))

        attr = torch.stack(multipliers, dim=0)
        if not raw_outputs:
            attr = attr.mean(dim=0)
            if not hypothetical:
                attr = attr * _X.cpu()
        
        attributions.append(attr)


    model.apply(_clear_hooks)
    for module in model.modules():
        del(module._NON_LINEAR_OPS)

    attributions = torch.stack(attributions).detach()

    if return_references:
        return attributions, torch.stack(references_).detach()
    return attributions



# space.py
# Contact: Jacob Schreiber <jmschreiber91@gmail.com>

import numpy
import torch

from .utils import _validate_input
from .utils import _cast_as_tensor
from .utils import one_hot_encode

from .ersatz import multisubstitute
from .predict import predict

from tqdm import tqdm


def space(model, X, motifs, spacing, start=None, alphabet=['A', 'C', 'G', 'T'], 
	func=predict, additional_func_kwargs={}, verbose=False, **kwargs):
	"""Runs a single spacing experiment and returns predictions.

	Given a predictive model, a set of motifs to insert and the spacings
	between them, and a set of background sequences, return the predictions 
	from the model when using the background sequences and after inserting 
	the motifs into the sequences.


	Parameters
	----------
	model: torch.nn.Module
		A PyTorch model to use for making predictions. These models can take in
		any number of inputs and make any number of outputs. The additional
		inputs must be specified in the `args` parameter.

	X: torch.tensor, shape=(-1, len(alphabet), length)
		A one-hot encoded set of sequences to have a motif inserted into.

	motifs: list of torch.tensor, shape=(-1, len(alphabet), motif_length)
		A list of strings or of one-hot encoded version of a short motif to 
		substitute into the set of sequences.

	spacing: torch.Tensor, shape=(-1, len(motifs)-1)
		A tensor specifying all spacings between motifs to consider. Each row
		in this tensor is a different combination of spacings between motifs
		and each column is the spacing between an adjacent pair of motifs.
		Specifically, the 1st column corresponds to the spacing between the
		first and second motif, the 2nd column corresponds to the spacing
		between the second and third motif, etc. 

	start: int or None, optional
		The starting position of where to insert the motif. If None, insert the
		motif into the middle of the sequence such that the middle of the motif
		occurs at the middle of the sequence. Default is None.

	alphabet : set or tuple or list, optional
		A pre-defined alphabet where the ordering of the symbols is the same
		as the index into the returned tensor, i.e., for the alphabet ['A', 'B']
		the returned tensor will have a 1 at index 0 if the character was 'A'.
		Characters outside the alphabet are ignored and none of the indexes are
		set to 1. This is not necessary or used if a one-hot encoded tensor is
		provided for the motif. Default is ['A', 'C', 'G', 'T'].

	func: function, optional
		A function to apply before and after making the substitutions. Default 
		is `predict`.

	additional_func_kwargs: dict, optional
		Additional named arguments to pass into the function when it is called.
		This is provided as an alternate path to route arguments into the 
		function in case they overlap, name-wise, with those in this function,
		or if you want to be absolutely sure that the arguments are making
		their way into the function. Default is {}.

	verbose: bool, optional
		Whether to display a progress bar as spacings are evaluated. Default
		is False.

	kwargs: optional
		Additional named arguments that will get passed into the function when
		it is called. Default is no arguments are passed in.


	Returns
	-------
	y_befores: torch.Tensor or list of torch.Tensors
		The predictions from the model before inserting the motif in. If the
		output from the model's forward function is a single tensor, it will
		return that. If the model outputs a list of tensors, it will return
		those.

	y_afters: torch.Tensor or list of torch.Tensors
		The predictions from the model after inserting the motif in. If the
		output from the model's forward function is a single tensor, it will
		return that. If the model outputs a list of tensors, it will return
		those.
	"""

	spacing = _validate_input(_cast_as_tensor(spacing, dtype=torch.int32), 
		"spacing", shape=(-1, len(motifs)-1))
	X = _validate_input(X, "X", shape=(-1, len(alphabet), -1))

	y_before = func(model, X, **kwargs, **additional_func_kwargs)
	y_afters = []

	for _spacing in tqdm(spacing, disable=not verbose):
		_spacing = [s.item() for s in _spacing]
		
		X_perturb = multisubstitute(X, motifs, _spacing, start=start, 
			alphabet=alphabet)

		y_after = func(model, X_perturb, **kwargs, **additional_func_kwargs)
		y_afters.append(y_after)

	if isinstance(y_afters[0], torch.Tensor):
		y_afters = torch.stack(y_afters).transpose(0, 1)
	else:
		y_afters = [torch.stack(y_).transpose(0, 1) for y_ in list(zip(
			*y_afters))]

	return y_before, y_afters



# test_match.py
# Contact: Jacob Schreiber <jmschreiber91@gmail.com>

import numpy
import torch
import pandas
import pytest

from tangermeme.utils import characters
from tangermeme.utils import one_hot_encode
from tangermeme.utils import random_one_hot

from tangermeme.io import extract_loci

from tangermeme.match import _calculate_char_perc
from tangermeme.match import _extract_and_filter_chrom
from tangermeme.match import extract_matching_loci

from numpy.testing import assert_raises
from numpy.testing import assert_array_almost_equal



###


def test_calculate_char_perc():
	seq = 'CAGCTCTAACTGATACTATCGAT'

	gc = _calculate_char_perc(seq, width=5, chars=['C', 'G'])
	at = _calculate_char_perc(seq, width=5, chars=['A', 'T'])
	n = _calculate_char_perc(seq, width=5, chars=['N'])

	assert gc.shape[0] == 4
	assert gc.dtype == numpy.float64

	assert_array_almost_equal(gc, [0.6, 0.4, 0.2, 0.4])
	assert_array_almost_equal(at, [0.4, 0.6, 0.8, 0.6])
	assert_array_almost_equal(gc+at, [1.0, 1.0, 1.0, 1.0])
	assert_array_almost_equal(n, [0.0, 0.0, 0.0, 0.0])


def test_calculate_char_perc_char():
	seq = 'ATCGATAACTACTACTACTGACGT'
	a = _calculate_char_perc(seq, width=5, chars='A')
	assert_array_almost_equal(a, [0.4, 0.4, 0.4, 0.2])


def test_calculate_char_perc_homopolymer():
	seq = 'CCCCCCCCCCCCCCCCCCCCCCC'

	gc = _calculate_char_perc(seq, width=5, chars=['C', 'G'])
	assert_array_almost_equal(gc, [1.0, 1.0, 1.0, 1.0])


def test_calculate_char_perc_N():
	seq = 'ACTATATGACACTCAGTAGCTNNNNNNNNCATCATACCATTACGACGTTCAAC'

	n = _calculate_char_perc(seq, width=10, chars=['N'])
	assert_array_almost_equal(n, [0. , 0. , 0.8, 0. , 0.])

	n = _calculate_char_perc(seq, width=10, chars='N')
	assert_array_almost_equal(n, [0. , 0. , 0.8, 0. , 0.])


def test_calculate_char_perc_short():
	seq = 'ATCGATACGT'

	gc = _calculate_char_perc(seq, width=10, chars=['C', 'G'])
	at = _calculate_char_perc(seq, width=10, chars=['A', 'T'])
	assert_array_almost_equal(gc, [0.4])
	assert_array_almost_equal(at, [0.6])


###


def test_extract_and_filter_chrom():
	regions = _extract_and_filter_chrom("tests/data/test.fa", 
		chrom='chr1', in_window=10, out_window=10)

	assert isinstance(regions, dict)
	assert len(regions) == 2
	assert_array_almost_equal(regions[25], [0, 1, 3, 5, 6, 7, 8, 10, 11, 12, 
		13, 14, 18, 22, 25, 26, 27])
	assert_array_almost_equal(regions[20], [2, 4, 9, 15, 16, 17, 19, 20, 21, 
		23, 24])


	regions = _extract_and_filter_chrom("tests/data/test.fa", 
		chrom='chr1', in_window=20, out_window=20)

	assert isinstance(regions, dict)
	assert len(regions) == 3
	assert_array_almost_equal(regions[25], [0, 3, 5, 6, 13])
	assert_array_almost_equal(regions[23], [1, 2, 4, 7, 9, 11, 12])
	assert_array_almost_equal(regions[20], [8, 10])


def test_extract_and_filter_chrom_gc_content():
	regions = _extract_and_filter_chrom("tests/data/test.fa", 
		chrom='chr1', in_window=20, out_window=20, gc_bin_width=0.05)

	for key, values in regions.items():
		chroms = ['chr1']*len(values)
		start = numpy.array(values)*20
		end = (numpy.array(values)+1)*20

		df = pandas.DataFrame({'chrom': chroms, 'start': start, 'end': end})
		X = extract_loci(df, "tests/data/test.fa", in_window=20)
		X = X.type(torch.float32)

		assert_array_almost_equal(X[:, [1, 2]].mean(axis=-1).sum(axis=1), 
			[key * 0.05]*X.shape[0])


def test_extract_and_filter_chrom_gc_content_large():
	regions = _extract_and_filter_chrom("tests/data/test.fa", 
		chrom='chr7', in_window=50, out_window=50)

	for key, values in regions.items():
		chroms = ['chr7']*len(values)
		start = numpy.array(values)*50
		end = (numpy.array(values)+1)*50

		df = pandas.DataFrame({'chrom': chroms, 'start': start, 'end': end})

		try:
			X = extract_loci(df, "tests/data/test.fa", in_window=50)
			X = X.type(torch.float32)
		except:
			continue

		assert_array_almost_equal(X[:, [1, 2]].mean(axis=-1).sum(axis=1), 
			[key * 0.02]*X.shape[0])


def test_extract_and_filter_chrom_some_N():
	regions = _extract_and_filter_chrom("tests/data/test.fa", 
		chrom='chr5', in_window=10, out_window=10)

	assert isinstance(regions, dict)
	assert len(regions) == 5
	assert_array_almost_equal(regions[30], [9, 14])
	assert_array_almost_equal(regions[25], [0, 4, 8])
	assert_array_almost_equal(regions[20], [2, 3, 15])
	assert_array_almost_equal(regions[15], [1, 7, 10, 11, 12])
	assert_array_almost_equal(regions[10], [13])


def test_extract_and_filter_chrom_many_N():
	regions = _extract_and_filter_chrom("tests/data/test.fa", 
		chrom='chr4', in_window=10, out_window=10)

	assert isinstance(regions, dict)
	assert len(regions) == 5
	assert_array_almost_equal(regions[15], [1, 7, 8])
	assert_array_almost_equal(regions[30], [2])
	assert_array_almost_equal(regions[20], [3, 20])
	assert_array_almost_equal(regions[25], [4, 5, 6, 16, 21, 22])
	assert_array_almost_equal(regions[10], [23])


def test_extract_and_filter_chrom_higher_N_filter():
	regions = _extract_and_filter_chrom("tests/data/test.fa", 
		chrom='chr4', in_window=10, out_window=10, max_n_perc=0.2)

	assert isinstance(regions, dict)
	assert len(regions) == 5
	assert_array_almost_equal(regions[15], [1, 7, 8, 15])
	assert_array_almost_equal(regions[30], [2])
	assert_array_almost_equal(regions[20], [3, 20])
	assert_array_almost_equal(regions[25], [4, 5, 6, 16, 21, 22])
	assert_array_almost_equal(regions[10], [23])


def test_extract_and_filter_chrom_no_N_filter():
	regions = _extract_and_filter_chrom("tests/data/test.fa", 
		chrom='chr4', in_window=10, out_window=10, max_n_perc=1)

	assert isinstance(regions, dict)
	assert len(regions) == 7
	assert_array_almost_equal(regions[15], [0, 1, 7, 8, 15])
	assert_array_almost_equal(regions[30], [2])
	assert_array_almost_equal(regions[20], [3, 20])
	assert_array_almost_equal(regions[25], [4, 5, 6, 16, 21, 22])
	assert_array_almost_equal(regions[10], [23])
	assert_array_almost_equal(regions[5], [17])
	assert_array_almost_equal(regions[0], [9, 10, 11, 12, 13, 14, 18, 19])


def test_extract_and_filter_chrom_N_in_windows_even():
	for window in range(10, 81, 10):
		regions = _extract_and_filter_chrom("tests/data/test.fa", 
			chrom='chr7', in_window=window, out_window=10, 
			gc_bin_width=1./window)

		for key, values in regions.items():
			chroms = ['chr7']*len(values)
			start = numpy.array(values)*window
			end = (numpy.array(values)+1)*window

			df = pandas.DataFrame({'chrom': chroms, 'start': start, 'end': end})

			try:
				X = extract_loci(df, "tests/data/test.fa", in_window=window)
				X = X.type(torch.float32)
			except:
				continue

			assert_array_almost_equal(X[:, [1, 2]].mean(axis=-1).sum(axis=1), 
				[key * 1./window]*X.shape[0])


def test_extract_and_filter_chrom_N_in_windows_odd():
	for window in range(10, 81, 5):
		regions = _extract_and_filter_chrom("tests/data/test.fa", 
			chrom='chr7', in_window=window, out_window=10, 
			gc_bin_width=1./window)

		for key, values in regions.items():
			chroms = ['chr7']*len(values)
			start = numpy.array(values)*window
			end = (numpy.array(values)+1)*window

			df = pandas.DataFrame({'chrom': chroms, 'start': start, 'end': end})

			try:
				X = extract_loci(df, "tests/data/test.fa", in_window=window)
				X = X.type(torch.float32)
			except:
				continue
				
			assert_array_almost_equal(X[:, [1, 2]].mean(axis=-1).sum(axis=1), 
				[key * 1./window]*X.shape[0])


def test_extract_and_filter_chrom_N_out_windows():
	for window in range(10, 81, 10):
		regions = _extract_and_filter_chrom("tests/data/test.fa", 
			chrom='chr7', in_window=10, out_window=window, gc_bin_width=0.1)

		for key, values in regions.items():
			chroms = ['chr7']*len(values)
			start = numpy.array(values)*10
			end = (numpy.array(values)+1)*10

			df = pandas.DataFrame({'chrom': chroms, 'start': start, 'end': end})

			try:
				X = extract_loci(df, "tests/data/test.fa", in_window=10)
				X = X.type(torch.float32)
			except:
				continue

			assert_array_almost_equal(X[:, [1, 2]].mean(axis=-1).sum(axis=1), 
				[key * 0.1]*X.shape[0])


def test_extract_and_filter_chrom_signal_threshold():
	regions = _extract_and_filter_chrom("tests/data/test.fa", 
		chrom='chr1', in_window=20, out_window=18, gc_bin_width=1.1,
		bigwig="tests/data/test.bw", signal_threshold=30)

	for key, values in regions.items():
		chroms = ['chr1']*len(values)
		start = numpy.array(values)*20
		end = (numpy.array(values)+1)*20

		df = pandas.DataFrame({'chrom': chroms, 'start': start, 'end': end})
		_, y = extract_loci(df, "tests/data/test.fa", ["tests/data/test.bw"], 
			in_window=20, out_window=18)

		assert len(y) == 14
		assert all(y.sum(axis=-1) <= 30)


	regions = _extract_and_filter_chrom("tests/data/test.fa", 
		chrom='chr1', in_window=20, out_window=18, gc_bin_width=1.1,
		bigwig="tests/data/test.bw", signal_threshold=10)

	for key, values in regions.items():
		chroms = ['chr1']*len(values)
		start = numpy.array(values)*20
		end = (numpy.array(values)+1)*20

		df = pandas.DataFrame({'chrom': chroms, 'start': start, 'end': end})
		_, y = extract_loci(df, "tests/data/test.fa", ["tests/data/test.bw"], 
			in_window=20, out_window=18)
		
		assert len(y) == 9
		assert all(y.sum(axis=-1) <= 15)


###


def test_extract_matching_loci():
	regions = extract_matching_loci("tests/data/test.bed", "tests/data/test.fa", 
		chroms=['chr1'], in_window=10, out_window=10, random_state=0)

	assert isinstance(regions, pandas.DataFrame)
	assert regions.shape == (3, 3)
	assert tuple(regions.columns) == ('chrom', 'start', 'end')

	assert numpy.unique(regions['chrom']).shape[0] == 1
	assert numpy.unique(regions['chrom']) == ('chr1',)
	assert tuple(regions['start']) == (70, 190, 240)
	assert tuple(regions['end']) == (80, 200, 250)

	X0 = extract_loci("tests/data/test.bed", "tests/data/test.fa", chroms=['chr1'], in_window=10)
	X1 = extract_loci(regions, "tests/data/test.fa", in_window=10)

	assert X0[:, [1, 2]].sum() == 13
	assert X1[:, [1, 2]].sum() == 13


def test_extract_matching_loci_start_edge():
	peaks = pandas.DataFrame({
		'chrom': ['chr1', 'chr1', 'chr1'],
		'start': [0, 15, 30],
		'end': [10, 25, 40]
	})

	regions = extract_matching_loci(peaks, "tests/data/test.fa", 
		chroms=['chr1'], in_window=10, out_window=10, random_state=0)
	assert regions.shape == (3, 3)


	regions = extract_matching_loci(peaks, "tests/data/test.fa", 
		chroms=['chr1'], in_window=15, out_window=10, random_state=0)
	assert regions.shape == (2, 3)


def test_extract_matching_loci_end_edge():
	peaks = pandas.DataFrame({
		'chrom': ['chr1', 'chr1', 'chr1'],
		'start': [260, 270, 280],
		'end': [270, 280, 290]
	})

	regions = extract_matching_loci(peaks, "tests/data/test.fa", 
		chroms=['chr1'], in_window=10, out_window=10, random_state=0)
	assert regions.shape == (2, 3)


	regions = extract_matching_loci(peaks, "tests/data/test.fa", 
		chroms=['chr1'], in_window=20, out_window=10, random_state=0)
	assert regions.shape == (1, 3)


def test_extract_matching_loci_some_N():
	peaks = pandas.DataFrame({
		'chrom': ['chr4', 'chr4', 'chr4'],
		'start': [0, 15, 30],
		'end': [10, 15, 30]
	})

	regions = extract_matching_loci(peaks, "tests/data/test.fa", 
		chroms=['chr4'], in_window=10, out_window=10, random_state=0)

	assert isinstance(regions, pandas.DataFrame)
	assert regions.shape == (2, 3)
	assert tuple(regions.columns) == ('chrom', 'start', 'end')

	assert numpy.unique(regions['chrom']).shape[0] == 1
	assert numpy.unique(regions['chrom']) == ('chr4',)
	assert tuple(regions['start']) == (20, 80)
	assert tuple(regions['end']) == (30, 90)

	X0 = extract_loci(peaks, "tests/data/test.fa", in_window=10)
	X1 = extract_loci(regions, "tests/data/test.fa", in_window=10)

	assert X0[:, [1, 2]].sum() == 12
	assert X1[:, [1, 2]].sum() == 9


def test_extract_matching_loci_N():
	peaks = pandas.DataFrame({
		'chrom': ['chr4', 'chr4', 'chr4'],
		'start': [0, 90, 100],
		'end': [10, 100, 110]
	})

	regions = extract_matching_loci(peaks, "tests/data/test.fa", 
		chroms=['chr4'], in_window=10, out_window=10, random_state=0)

	assert isinstance(regions, pandas.DataFrame)
	assert regions.shape == (0, 3)
	assert tuple(regions.columns) == ('chrom', 'start', 'end')

	assert tuple(regions['chrom']) == tuple()
	assert tuple(regions['start']) == tuple()
	assert tuple(regions['end']) == tuple()


def test_extract_matching_loci_allow_N():
	peaks = pandas.DataFrame({
		'chrom': ['chr4', 'chr4', 'chr4'],
		'start': [0, 90, 100],
		'end': [10, 100, 110]
	})

	regions = extract_matching_loci(peaks, "tests/data/test.fa", 
		chroms=['chr4'], in_window=10, out_window=10, max_n_perc=1.1, 
		random_state=0)

	assert isinstance(regions, pandas.DataFrame)
	assert regions.shape == (3, 3)
	assert tuple(regions.columns) == ('chrom', 'start', 'end')

	assert tuple(regions['chrom']) == ('chr4', 'chr4', 'chr4')
	assert tuple(regions['start']) == (70, 120, 140)
	assert tuple(regions['end']) == (80, 130, 150)
